"""Advanced rocket demo featuring grid fins and exploded view."""

from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Tuple

from yapcad.geom import point, vect
from yapcad.geom3d import solid, surface, translatesolid, rotatesolid
from yapcad.geom3d_util import conic, extrude
from yapcad.io import write_stl, write_step


def merge_solids(parts: List[list]) -> list:
    surfaces: List = []
    for part in parts:
        surfaces.extend(part[1])
    return solid(surfaces)


def stack_conic(radius_base: float, radius_top: float, height: float, base_z: float) -> list:
    return conic(radius_base, radius_top, height, center=point(0, 0, base_z))


def create_engine(x: float, y: float, base_z: float, bell_radius: float, height: float) -> list:
    bell = conic(bell_radius, bell_radius * 0.4, height, center=point(x, y, base_z))
    nozzle = conic(bell_radius * 0.4, bell_radius * 0.4, height * 0.2, center=point(x, y, base_z + height))
    return merge_solids([bell, nozzle])


def create_box(width: float, depth: float, height: float, center: Tuple[float, float, float]) -> list:
    cx, cy, cz = center
    hx, hy, hz = width / 2.0, depth / 2.0, height / 2.0
    base_z = cz - hz
    poly = [
        point(cx - hx, cy - hy, base_z),
        point(cx + hx, cy - hy, base_z),
        point(cx + hx, cy + hy, base_z),
        point(cx - hx, cy + hy, base_z),
    ]
    normals = [[0, 0, 1, 0] for _ in poly]
    faces = [[0, 1, 2], [0, 2, 3]]
    boundary = [0, 1, 2, 3]
    base_surface = surface(poly, normals, faces, boundary, [])
    return extrude(base_surface, height, direction=vect(0, 0, 1, 0))


def create_grid_fin_panel(width: float, height: float, thickness: float, cells: int) -> list:
    bars: List[list] = []
    half_w, half_h = width / 2.0, height / 2.0

    frame = thickness

    # Outer frame top/bottom
    bars.append(create_box(thickness, width, thickness, (0.0, 0.0, half_h - thickness / 2.0)))
    bars.append(create_box(thickness, width, thickness, (0.0, 0.0, -half_h + thickness / 2.0)))

    # Outer frame sides
    bars.append(create_box(thickness, thickness, height, (0.0, -half_w + thickness / 2.0, 0.0)))
    bars.append(create_box(thickness, thickness, height, (0.0, half_w - thickness / 2.0, 0.0)))

    # Internal horizontal bars
    if cells > 1:
        step = height / cells
        for i in range(1, cells):
            z = -half_h + i * step
            bars.append(create_box(thickness, width - 2 * frame, thickness / 1.5, (0.0, 0.0, z)))

    # Internal vertical bars
    if cells > 1:
        step = width / cells
        for i in range(1, cells):
            y = -half_w + i * step
            bars.append(create_box(thickness, thickness / 1.5, height - 2 * frame, (0.0, y, 0.0)))

    return merge_solids(bars)


def build_components() -> Dict[str, list]:
    components: Dict[str, list] = {}

    booster_height = 18.0
    booster_radius = 2.6
    interstage_height = 3.0
    stage2_height = 8.0
    fairing_lower_height = 2.0
    fairing_upper_height = 2.5

    booster_body = stack_conic(booster_radius, booster_radius, booster_height, 0.0)

    # Engine cluster
    engines = [create_engine(0.0, 0.0, -2.5, 0.55, 2.2)]
    ring = 1.4
    for dx, dy in [(ring, 0.0), (-ring, 0.0), (0.0, ring), (0.0, -ring)]:
        engines.append(create_engine(dx, dy, -2.5, 0.45, 1.8))
    engine_cluster = merge_solids(engines)

    # Grid fins near top of booster
    fin_panel = create_grid_fin_panel(width=3.4, height=3.0, thickness=0.35, cells=4)
    fin_z = 14.0
    fin_offset = booster_radius + 0.25
    grid_fins: List[list] = []
    translated = translatesolid(fin_panel, point(fin_offset, 0.0, fin_z))
    grid_fins.append(translated)
    for angle in (90, 180, 270):
        grid_fins.append(rotatesolid(translated, angle, cent=point(0, 0, fin_z), axis=point(0, 0, 1)))
    grid_fin_cluster = merge_solids(grid_fins)

    booster = merge_solids([booster_body, engine_cluster, grid_fin_cluster])
    components['booster'] = booster

    interstage = stack_conic(booster_radius, 1.9, interstage_height, booster_height)
    components['interstage'] = interstage

    stage2 = stack_conic(1.9, 1.9, stage2_height, booster_height + interstage_height)
    upper_engine = create_engine(0.0, 0.0, booster_height + interstage_height - 1.5, 0.4, 1.6)
    payload_barrel = stack_conic(1.9, 1.9, 3.0, booster_height + interstage_height + stage2_height)
    fairing_lower = stack_conic(1.9, 1.1, fairing_lower_height, booster_height + interstage_height + stage2_height + 3.0)
    fairing_upper = stack_conic(1.1, 0.1, fairing_upper_height, booster_height + interstage_height + stage2_height + 3.0 + fairing_lower_height)

    upper_stage = merge_solids([stage2, payload_barrel, fairing_lower, fairing_upper, upper_engine])
    components['upper_stage'] = upper_stage

    return components


def build_assembly(components: Dict[str, list]) -> list:
    return merge_solids(list(components.values()))


def visualise_exploded(components: Dict[str, list]) -> None:
    try:
        from yapcad.pyglet_drawable import pygletDraw
    except Exception as exc:  # pragma: no cover - visual only
        print(f'Visualisation skipped: {exc}')
        return

    offsets = {
        'booster': 0.0,
        'interstage': 5.0,
        'upper_stage': 10.0,
    }

    palette = {
        'booster': 'silver',
        'interstage': 'olive',
        'upper_stage': 'navy',
    }

    dd = pygletDraw()
    for name, solid_obj in components.items():
        offset = offsets.get(name, 0.0)
        display_obj = translatesolid(solid_obj, point(0, 0, offset))
        dd.make_object(name, lighting=True, material='pearl')
        dd.linecolor = palette.get(name, 'gray')
        dd.draw(display_obj, name=name)
    dd.display()


def export_models(assembly: list, stem: str) -> None:
    stl_path = Path(f'{stem}.stl')
    step_path = Path(f'{stem}.step')
    write_stl(assembly, stl_path, binary=True, name=stem)
    write_step(assembly, step_path, name=stem)
    print(f'Exported STL: {stl_path.resolve()}')
    print(f'Exported STEP: {step_path.resolve()}')


def main() -> None:
    components = build_components()
    assembly = build_assembly(components)
    export_models(assembly, 'rocket_grid_demo')
    visualise_exploded(components)


if __name__ == '__main__':
    main()



import io

from yapcad.geom import point, vect
from yapcad.geom3d import poly2surfaceXY, surface, solid, translatesolid
from yapcad.geom3d_util import extrude
from yapcad.io.step import write_step


def _make_surface():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]
    surf, _ = poly2surfaceXY(poly)
    return surf


def test_write_step_basic():
    surf = _make_surface()
    buf = io.StringIO()
    write_step(surf, buf, name='test_step')
    data = buf.getvalue()
    assert data.startswith('ISO-10303-21;')
    assert 'OPEN_SHELL' in data
    assert 'SHELL_BASED_SURFACE_MODEL' in data
    assert 'MANIFOLD_SOLID_BREP' not in data
    assert 'ADVANCED_FACE' in data
    assert 'test_step' in data


def _make_box(origin=None):
    if origin is None:
        origin = (0.0, 0.0, 0.0)
    x0, y0, z0 = origin
    base = [
        point(x0, y0, z0),
        point(x0 + 1.0, y0, z0),
        point(x0 + 1.0, y0 + 1.0, z0),
        point(x0, y0 + 1.0, z0),
    ]
    normals = [vect(0, 0, 1, 0) for _ in base]
    faces = [[0, 1, 2], [0, 2, 3]]
    boundary = [0, 1, 2, 3]
    base_surface = surface(base, normals, faces, boundary, [])
    return extrude(base_surface, distance=1.0)


def test_write_step_solid_component():
    box = _make_box()
    buf = io.StringIO()
    write_step(box, buf, name='solid_box')
    data = buf.getvalue()
    assert 'MANIFOLD_SOLID_BREP' in data
    assert data.count('MANIFOLD_SOLID_BREP') == 1
    assert 'CLOSED_SHELL' in data
    assert 'OPEN_SHELL' not in data
    assert max(len(line) for line in data.splitlines()) <= 256


def test_write_step_multiple_components():
    box1 = _make_box()
    box2 = translatesolid(_make_box(), point(2.5, 0.0, 0.0))
    assembly = solid(box1[1] + box2[1])
    buf = io.StringIO()
    write_step(assembly, buf, name='assembly')
    data = buf.getvalue()
    assert data.count('MANIFOLD_SOLID_BREP') == 2
    assert 'SHELL_BASED_SURFACE_MODEL' not in data
    assert max(len(line) for line in data.splitlines()) <= 256



import math

from yapcad.geom import point
from yapcad.geom3d import poly2surfaceXY
from yapcad.geometry_utils import (
    Triangle,
    orient_triangle,
    to_point,
    to_vec3,
    triangle_area,
    triangle_centroid,
    triangle_is_degenerate,
    triangle_normal,
    triangles_from_mesh,
)
from yapcad.mesh import mesh_view


def test_to_vec3_and_to_point_roundtrip():
    pt = point(1.2, -3.4, 5.6)
    vec = to_vec3(pt)
    assert vec == (1.2, -3.4, 5.6)
    lifted = to_point(vec)
    assert lifted == (1.2, -3.4, 5.6, 1.0)


def test_triangle_normal_and_area():
    v0 = (0.0, 0.0, 0.0)
    v1 = (1.0, 0.0, 0.0)
    v2 = (0.0, 1.0, 0.0)

    n = triangle_normal(v0, v1, v2)
    assert n == (0.0, 0.0, 1.0)
    assert math.isclose(triangle_area(v0, v1, v2), 0.5)


def test_triangle_is_degenerate_when_colinear():
    v0 = (0.0, 0.0, 0.0)
    v1 = (1.0, 1.0, 1.0)
    v2 = (2.0, 2.0, 2.0)
    assert triangle_is_degenerate(v0, v1, v2)
    assert triangle_normal(v0, v1, v2) is None


def test_orient_triangle_flips_for_reference():
    v0 = (0.0, 0.0, 0.0)
    v1 = (1.0, 0.0, 0.0)
    v2 = (0.0, 1.0, 0.0)

    oriented = orient_triangle(v0, v1, v2, preferred_normal=(0.0, 0.0, -1.0))
    assert oriented[1] == v2 and oriented[2] == v1


def test_triangle_centroid():
    v0 = (0.0, 0.0, 0.0)
    v1 = (3.0, 0.0, 0.0)
    v2 = (0.0, 3.0, 0.0)
    cx, cy, cz = triangle_centroid(v0, v1, v2)
    assert math.isclose(cx, 1.0)
    assert math.isclose(cy, 1.0)
    assert math.isclose(cz, 0.0)


def test_triangles_from_mesh_produces_dataclasses():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]
    surf, _ = poly2surfaceXY(poly)
    tris = list(triangles_from_mesh(mesh_view(surf)))
    assert len(tris) == 1
    assert isinstance(tris[0], Triangle)
    assert tris[0].normal == (0.0, 0.0, 1.0)



#!/usr/bin/env python3
"""Diagnose boolean meshes produced by yapCAD using trimesh helpers."""

from __future__ import annotations

import argparse
import json
from pathlib import Path

from mesh_diagnostics import MeshDiagnostics, diagnose_mesh, export_boundary

from yapcad.geom import point
from yapcad.geom3d import solid_boolean, translatesolid
from yapcad.geom3d_util import prism, sphere, conic, tube

from yapcad.boolean.trimesh_engine import _solid_to_mesh


def _build_solids(kind: str):
    if kind == 'cube':
        a = prism(2, 2, 2)
        b = translatesolid(prism(2, 2, 2), point(0.75, 0.0, 0.0))
    elif kind == 'sphere':
        a = sphere(2.0)
        b = translatesolid(sphere(2.0), point(1.0, 0.0, 0.0))
    elif kind == 'box_hole':
        a = prism(2, 2, 2)
        b = conic(0.6, 0.6, 3.0, center=point(0.0, 0.0, -1.5))
    elif kind == 'tube_hole':
        a = tube(outer_diameter=3.0, wall_thickness=0.5, length=4.0,
                 base_point=point(0.0, 0.0, -2.0))
        b = conic(0.6, 0.6, 4.2, center=point(1.2, 0.0, -2.1))
    else:
        raise ValueError(f'unsupported shape kind: {kind!r}')
    return a, b


def main():
    parser = argparse.ArgumentParser(description='Diagnose boolean mesh quality.')
    parser.add_argument('--shapes', choices=['cube', 'sphere', 'box_hole', 'tube_hole'], required=True)
    parser.add_argument('--operation', choices=['union', 'intersection', 'difference'], default='union')
    parser.add_argument('--engine', default=None, help='boolean engine (e.g., native, trimesh:manifold)')
    parser.add_argument('--stitch', action='store_true', help='enable stitch flag for boolean call')
    parser.add_argument('--boundary-output', type=Path, help='optional path to export boundary edges (PLY)')
    parser.add_argument('--json', type=Path, help='write diagnostics JSON to path')
    args = parser.parse_args()

    solids = _build_solids(args.shapes)
    result = solid_boolean(solids[0], solids[1], args.operation,
                           stitch=args.stitch, engine=args.engine)

    mesh = _solid_to_mesh(result)
    diagnostics = diagnose_mesh(mesh)

    if args.boundary_output:
        export_boundary(mesh, args.boundary_output)

    print(json.dumps(diagnostics.to_dict(), indent=2))

    if args.json:
        args.json.parent.mkdir(parents=True, exist_ok=True)
        args.json.write_text(json.dumps(diagnostics.to_dict(), indent=2))


if __name__ == '__main__':
    main()



## yapCAD boolean operations geometry example

from yapcad.geom import *
from yapcad.poly import *
from yapcad.combine import *

import math
import random

def makeStars(center,rad=5.0,diam=2.0):
    poly1 = Polygon()
    poly2 = Polygon()

    for i in range(8):
        ang = (pi2/8)*i
        p = poly1
        if i%2 ==0:
            p = poly2
        p.addArc(arc(add(center, point(math.cos(ang)*rad,
                                       math.sin(ang)*rad)),diam))
    return poly1,poly2

def geometry():
    p1 = Polygon()
    x = point(-2,2)
    p1.addArc(arc(x,3.0))

    #p1 = RoundRect(4,4,0.2,point(-4.0,4.0))

    p2 = RoundRect(7,6,1,point(1,-1))
    b = Boolean('union',[p1,p2])

    p4,p5 = makeStars(point(12,-10))
    b2 = Boolean('union',[p4,p5])

    p6,p7 = makeStars(point(12,10))
    b3 = Boolean('intersection',[p6,p7])

    p8,p9 = makeStars(point(-12,-10))
    b4 = Boolean('difference',[p9,p8])

    p10,p11 = makeStars(point(-12,10))

    return [p1.geom,p2.geom,b.geom,b2.geom,b3.geom,b4.geom,p10.geom,p11.geom]
    #return [p1.geom,p2.geom,b.geom]

def testAndDraw(dd):
    dd.linecolor = 'red'

    gl = geometry()

    g1 = gl[0]
    g2 = gl[1]
    g3 = gl[2]
    g4 = gl[3]
    g5 = gl[4]
    g6 = gl[5]
    
    dd.linecolor = 'aqua'

    dd.draw(g3)

    g3r = rotate(g3,45)

    dd.linecolor = 'red'
    dd.draw(g3r)

    dd.linecolor = 'white'
    for i in range(len(g4)):
        dd.linecolor = i%7+1
        dd.draw(g4[i])
    
    dd.linecolor = 'silver'
    dd.draw(translate(g4,point(0,0,-1)))
            
    dd.linecolor = 'white'
    dd.draw(g5)

    dd.linecolor = 'yellow'
    dd.draw(g6)

    dd.linecolor = 'aqua'
    dd.draw(gl[6:])

    dd.display()
    
if __name__ == "__main__":
    import sys
    renderOgl = False
    filename="example11-out"
    oglarg= ("pyglet","opengl","OpenGL")
    dxfarg= ("ezdxf","dxf","DXF")
    print("example11.py -- yapCAD boolean operations demonstration")
    if len(sys.argv) > 1:
        if sys.argv[1] in oglarg:
            renderOgl=True
        elif sys.argv[1] in dxfarg:
            renderOgl=False
        else:
            print(" In this example, we create yapCAD Polygon() geometry generators and")
            print(" perform boolean operations on them to create combined figures.")
            print("syntax: $ python3 {} <rendertype> [filename]".format(sys.argv[0]))
            print("    where <rendertype> is one of {} for OpenGL".format(oglarg))
            print("    or one of {} for DXF".format(dxfarg))
            print("    For DXF, you can optionally specify [filename].dxf as the output file")
            quit()
    if len(sys.argv) > 2 and renderOgl==False:
        filename = sys.argv[2]+".dxf"
    requestedOgl = renderOgl
    dd = None
    if renderOgl:
        print("OpenGL rendering selected")
        try:
            from yapcad.pyglet_drawable import *
            dd=pygletDraw()
        except RuntimeError as err:
            print("\nSkipping OpenGL rendering: {}".format(err))
            renderOgl = False
        except Exception as err:
            print("\nSkipping OpenGL rendering due to unexpected error: {}".format(err))
            renderOgl = False

    if not renderOgl or dd is None:
        if requestedOgl and dd is None:
            print("Falling back to DXF rendering")
        else:
            print("DXF rendering selected")
        from yapcad.ezdxf_drawable import *
        #set up DXF rendering
        dd=ezdxfDraw()
        dd.filename = filename
        renderOgl = False
    print("rendering...")
    testAndDraw(dd)
    print("done")



import math

from yapcad.combine import Boolean
from yapcad.geom import arc, point
from yapcad.geom3d import solidbbox
from yapcad.geom3d_util import tube, conic_tube, spherical_shell, stack_solids
from yapcad.geom_util import geomlist2poly_components
from yapcad.poly import Polygon


def _radius(point_xyz, center_xy):
    dx = point_xyz[0] - center_xy[0]
    dy = point_xyz[1] - center_xy[1]
    return math.hypot(dx, dy)


def _surface_vertices(surface):
    return surface[1]


def _surface_normals(surface):
    return surface[2]


def test_boolean_difference_handles_arcs():
    outer = Polygon()
    outer.addArc(arc(point(0, 0), 10.0))
    hole = Polygon()
    hole.addArc(arc(point(0, 0), 2.0))

    shape = Boolean('difference', [outer, hole])
    outer_poly, holes = geomlist2poly_components(shape.geom)[0]
    assert len(outer_poly) > 3
    assert len(holes) == 1


def test_tube_basic_properties():
    sld = tube(outer_diameter=10.0, wall_thickness=2.0, length=8.0)
    assert sld[0] == 'solid'
    assert len(sld[1]) == 4

    bbox = solidbbox(sld)
    outer_radius = 5.0
    assert math.isclose(bbox[0][2], 0.0, abs_tol=1e-6)
    assert math.isclose(bbox[1][2], 8.0, abs_tol=1e-6)
    assert math.isclose(bbox[1][0], outer_radius, abs_tol=1e-4)

    outer_side = sld[1][1]
    verts = _surface_vertices(outer_side)
    norms = _surface_normals(outer_side)
    sample_idx = 0
    radius_vec = (verts[sample_idx][0], verts[sample_idx][1])
    normal = norms[sample_idx]
    assert radius_vec[0] * normal[0] + radius_vec[1] * normal[1] > 0

    inner_side = sld[1][3]
    verts_inner = _surface_vertices(inner_side)
    norms_inner = _surface_normals(inner_side)
    center = (0.0, 0.0)
    vec_to_center = (center[0] - verts_inner[0][0], center[1] - verts_inner[0][1])
    dot_inner = vec_to_center[0] * norms_inner[0][0] + vec_to_center[1] * norms_inner[0][1]
    assert dot_inner > 0


def test_conic_tube_tapers():
    sld = conic_tube(bottom_outer_diameter=12.0,
                     top_outer_diameter=8.0,
                     wall_thickness=1.5,
                     length=6.0)
    bbox = solidbbox(sld)
    assert math.isclose(bbox[0][2], 0.0, abs_tol=1e-6)
    assert math.isclose(bbox[1][2], 6.0, abs_tol=1e-6)

    top_surface = sld[1][2]
    center_xy = (0.0, 0.0)
    radii = [_radius(v[:3], center_xy) for v in _surface_vertices(top_surface)]
    assert math.isclose(max(radii), 4.0, rel_tol=1e-2)


def test_spherical_shell_full():
    sld = spherical_shell(outer_diameter=10.0, wall_thickness=1.0)
    assert len(sld[1]) == 2
    outer_surface = sld[1][0]
    inner_surface = sld[1][1]
    center = (0.0, 0.0, 0.0)
    sample_outer = _surface_vertices(outer_surface)[0]
    normal_outer = _surface_normals(outer_surface)[0]
    vec_outer = (sample_outer[0] - center[0], sample_outer[1] - center[1], sample_outer[2] - center[2])
    assert vec_outer[0] * normal_outer[0] + vec_outer[1] * normal_outer[1] + vec_outer[2] * normal_outer[2] > 0

    sample_inner = _surface_vertices(inner_surface)[0]
    normal_inner = _surface_normals(inner_surface)[0]
    vec_inner = (center[0] - sample_inner[0], center[1] - sample_inner[1], center[2] - sample_inner[2])
    assert vec_inner[0] * normal_inner[0] + vec_inner[1] * normal_inner[1] + vec_inner[2] * normal_inner[2] > 0


def test_spherical_shell_cap_has_conic_side():
    sld = spherical_shell(outer_diameter=10.0,
                          wall_thickness=1.0,
                          solid_angle=2 * math.pi)
    assert len(sld[1]) == 3
    conic_surface = sld[1][2]
    radii = {_radius(v[:3], (0.0, 0.0)) for v in _surface_vertices(conic_surface)}
    assert min(radii) < max(radii)


def test_stack_solids_spacing_directives():
    a = tube(outer_diameter=4.0, wall_thickness=0.5, length=2.0)
    b = tube(outer_diameter=4.0, wall_thickness=0.5, length=2.0)
    c, d = stack_solids(['space:3.0', a, 'space:1.5', b], axis='z', start=0.0, gap=0.0)

    bbox_c = solidbbox(c)
    bbox_d = solidbbox(d)

    assert math.isclose(bbox_c[0][2], 3.0, abs_tol=1e-6)
    assert math.isclose(bbox_d[0][2] - bbox_c[1][2], 1.5, abs_tol=1e-6)



import pytest
from yapcad.xform import *
## unit tests for yapCAD xform.py

class TestXform:
    """unit tests for yapCAD matrix operations"""

    def test_matrix(self):
        foo = Matrix([1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16])
        fooT = Matrix(foo,True)
        bar = Matrix([[1,0,0,1],[0,1,0,1],[0,0,1,1],[0,0,0,1]])
        baz = geom.vect(1,2,3)
        I = Matrix()
        a = 10.0
        assert(I.mul(bar).m == bar.m)
        assert(I.mul(foo).m == foo.m)
        assert(I.mul(fooT).m == fooT.mul(I).m)
        assert(I.mul(I).m == I.m)
        assert(foo.mul(bar).m == [[1,2,3,10],[5,6,7,26],[9,10,11,42],[13,14,15,58]])
        assert(foo.mul(baz) == [18, 46, 74, 102])
        assert(foo.mul(a).m == [[10.0,20.0,30.0,40.0],
                                [50.0,60.0,70.0,80.0],
                                [90.0,100.0,110.0,120.0],
                                [130.0,140.0,150.0,160.0]])
        assert(I.mul(baz) == baz)
        ## homogeneous coordinates test
        assert(geom.homo(foo.mul(baz)) ==
               [18.0/102.0, 46.0/102.0, 74.0/102.0, 1.0])

    def test_translation(self):
        p = geom.vect(1,2,3)
        dx = 4
        dy = -2
        dz = 1
        T = Translation([dx,dy,dz])
        assert(T.mul(p) == geom.vect(1+dx,2+dy,3+dz))
        Tinv = Translation([dx,dy,dz],inverse=True)
        assert(Tinv.mul(p) == geom.vect(1-dx,2-dy,3-dz))
    
    





#!/usr/bin/env python3

# Updated validation script using current pythonocc API

from pathlib import Path

from OCC.Core.STEPControl import STEPControl_Reader
from OCC.Core.IFSelect import IFSelect_RetDone

def describe_failures(reader: STEPControl_Reader) -> None:
      ws = reader.WS()
      model = ws.Model()

      for idx in range(1, model.NbEntities() + 1):
         entity = model.Value(idx)
         check = model.Check(entity)
         if check.HasFailed():
             print(f"Entity #{idx} failed: {check}")


def describe_failures_bad(reader: STEPControl_Reader) -> None:
    ws = reader.WS()
    model = ws.Model()
    stats = model.GlobalCheck()
    count = stats.NbFails()
    if count:
        print(f"Global check reported {count} failure(s):")
        for idx in range(1, count + 1):
            print(f"  - {stats.Fail(idx).Value().Message().ToCString()}")

    model = ws.Model()
    for idx in range(1, model.NbEntities() + 1):
        entity = model.Entity(idx)
        check = model.Check(entity)
        if check.HasFailed():
            print(f"Entity #{idx} failed: {check}")


def main(path: str) -> None:
    if not Path(path).exists():
        raise SystemExit(f"STEP file not found: {path}")

    reader = STEPControl_Reader()
    status = reader.ReadFile(path)

    if status != IFSelect_RetDone:
        print(f"ReadFile failed with status {status}")
        describe_failures(reader)
        return

    reader.TransferRoots()
    shape = reader.Shape()
    if shape.IsNull():
        print("Transfer succeeded, but resulting shape is null")

    describe_failures(reader)


if __name__ == '__main__':
    main('rocket_grid_demo.step')



"""Rocket internal layout demo using yapCAD parametric primitives.

Builds a single-stage liquid rocket cross-section with fuel/oxidizer tanks,
feed lines, turbopumps, and a simplified engine.  Outputs STL and STEP files
for inspection.  Geometry is generated headlessly; set ``YAPCAD_OUTPUT_DIR``
to control where artifacts land (defaults to the examples directory).
"""

from __future__ import annotations

import math
import os
from pathlib import Path

from yapcad.geom import epsilon, point, vect
from yapcad.geom3d import solid, solidbbox
from yapcad.geom3d_util import (
    conic_tube,
    rotatesolid,
    spherical_shell,
    stack_solids,
    translatesolid,
    tube,
)
from yapcad.io import write_step, write_stl


OUTPUT_DIR = Path(os.environ.get('YAPCAD_OUTPUT_DIR', 'examples'))
OUTPUT_DIR.mkdir(parents=True, exist_ok=True)


def merge_solids(solids):
    surfaces = []
    for sld in solids:
        surfaces.extend(sld[1])
    return solid(surfaces, [], ['procedure', 'rocket_internal_demo'])


def cutaway_solid_x(sld, plane_x=0.0, eps=1e-6):
    """Return a copy of ``sld`` with triangles strictly on the +X side removed."""

    stripped = []
    for surf in sld[1]:
        faces = []
        for tri in surf[3]:
            xs = [surf[1][idx][0] for idx in tri]
            if max(xs) <= plane_x + eps:
                faces.append(tri)
        if faces:
            stripped.append(['surface', surf[1], surf[2], faces, surf[4], surf[5]])
    return solid(stripped, [], sld[3] if len(sld) > 3 else [])


def make_tank(z_start, length, outer_radius, wall):
    cylinder_length = max(length - 2 * outer_radius, wall)
    cylinder_base = z_start + outer_radius
    body = tube(outer_diameter=outer_radius * 2,
                wall_thickness=wall,
                length=cylinder_length,
                base_point=point(0, 0, cylinder_base),
                include_caps=False)

    front_cap_center = point(0, 0, z_start + outer_radius)
    front_cap = spherical_shell(outer_diameter=outer_radius * 2,
                                wall_thickness=wall,
                                solid_angle=2 * math.pi,
                                center=front_cap_center)
    front_cap = rotatesolid(front_cap, 180, axis=point(0, 1, 0), cent=front_cap_center)

    back_cap_center = point(0, 0, z_start + outer_radius + cylinder_length)
    back_cap = spherical_shell(outer_diameter=outer_radius * 2,
                               wall_thickness=wall,
                               solid_angle=2 * math.pi,
                               center=back_cap_center)

    return merge_solids([body, front_cap, back_cap])


def make_engine(z_start):
    wall = 0.25
    chamber_d = 4.0
    chamber_l = 2.5
    throat_l = 1.5
    nozzle_l = 5.0

    nozzle = conic_tube(bottom_outer_diameter=6.0,
                        top_outer_diameter=2.2,
                        wall_thickness=wall,
                        length=nozzle_l,
                        base_point=point(0, 0, z_start),
                        include_caps=False)

    throat = tube(outer_diameter=2.2,
                  wall_thickness=wall,
                  length=throat_l,
                  base_point=point(0, 0, z_start + nozzle_l),
                  include_caps=False)

    chamber = conic_tube(bottom_outer_diameter=2.2,
                         top_outer_diameter=chamber_d,
                         wall_thickness=wall,
                         length=chamber_l,
                         base_point=point(0, 0, z_start + nozzle_l + throat_l),
                         include_caps=False)

    cap_center = z_start + nozzle_l + throat_l + chamber_l
    front_cap = spherical_shell(outer_diameter=chamber_d,
                                wall_thickness=wall,
                                solid_angle=2 * math.pi,
                                center=point(0, 0, cap_center))

    return merge_solids([nozzle, throat, chamber, front_cap])


def build_body_shell(cylinder_length, radius, wall):
    body = tube(outer_diameter=radius * 2,
                wall_thickness=wall,
                length=cylinder_length,
                center=point(0, 0, 0),
#                center=point(0, 0, cylinder_length / 2.0),
                include_caps=False)
    nose_center = point(0, 0, cylinder_length )
    nose = spherical_shell(outer_diameter=radius * 2,
                           wall_thickness=wall,
                           solid_angle=2 * math.pi,
                           center=nose_center)
    shell = merge_solids([body, nose])
    shell = cutaway_solid_x(shell, plane_x=0.0)
    bbox = solidbbox(shell)
    if bbox[0][2] > 0.0:
        shell = translatesolid(shell, vect(0, 0, -bbox[0][2], 0))
    return shell


def main():
    body_radius = 6.0
    wall = 0.3
    cylinder_length = 32.0
    body_shell = build_body_shell(cylinder_length, body_radius, wall)

    engine = make_engine(0.0)
    fuel_tank = make_tank(0.0, length=8.0, outer_radius=3.0, wall=0.25)
    oxidizer_tank = make_tank(0.0, length=9.0, outer_radius=3.8, wall=0.25)

    engine, oxidizer_tank, fuel_tank = stack_solids(
        [engine, 'space:2.0', oxidizer_tank, 'space:2.0', fuel_tank],
        axis='z',
        start=0.0,
        gap=0.0,
        align='center',
    )

    min_z = min(solidbbox(obj)[0][2] for obj in (engine, fuel_tank, oxidizer_tank))
    if abs(min_z) > epsilon:
        offset = -min_z
        engine = translatesolid(engine, vect(0, 0, offset, 0))
        fuel_tank = translatesolid(fuel_tank, vect(0, 0, offset, 0))
        oxidizer_tank = translatesolid(oxidizer_tank, vect(0, 0, offset, 0))

    top_target = cylinder_length - 2.0
    current_top = max(solidbbox(obj)[1][2] for obj in (fuel_tank, oxidizer_tank))
    adjust = top_target - current_top
    engine = translatesolid(engine, vect(0, 0, adjust, 0))
    fuel_tank = translatesolid(fuel_tank, vect(0, 0, adjust, 0))
    oxidizer_tank = translatesolid(oxidizer_tank, vect(0, 0, adjust, 0))

    min_z = min(solidbbox(obj)[0][2] for obj in (engine, fuel_tank, oxidizer_tank))
    if min_z < 0.0:
        engine = translatesolid(engine, vect(0, 0, -min_z, 0))
        fuel_tank = translatesolid(fuel_tank, vect(0, 0, -min_z, 0))
        oxidizer_tank = translatesolid(oxidizer_tank, vect(0, 0, -min_z, 0))

    engine = translatesolid(engine, vect(0, 0, -4.0, 0))

    components = [body_shell, engine, fuel_tank, oxidizer_tank]

    assembly = merge_solids(components)
    bbox = solidbbox(assembly)
    print('Bounding box:', bbox)

    stl_path = OUTPUT_DIR / 'rocket_internal_demo.stl'
    step_path = OUTPUT_DIR / 'rocket_internal_demo.step'
    write_stl(assembly, stl_path, binary=True, name='rocket_internal_demo')
    write_step(assembly, step_path, name='rocket_internal_demo')
    print('Exported STL to', stl_path)
    print('Exported STEP to', step_path)


if __name__ == '__main__':
    main()



## geom3d_util, additional 3D geometry support for yapCAD
## started on Mon Feb 15 20:23:57 PST 2021 Richard W. DeVaul

from yapcad.geom import *
from yapcad.geom_util import *
from yapcad.xform import *
from yapcad.geom3d import *

import math

"""
==================================================
Utility functions to support 3D geometry in yapCAD
==================================================

This module is mostly a collection of 
parametric soids and surfaces, and supporting functions.

"""


def sphere2cartesian(lat,lon,rad):
    """
    Utility function to convert spherical polar coordinates to
    cartesian coordinates for a sphere centered at the origin.
        ``lat`` -- latitude
        ``lon`` -- longitude
        ``rad`` -- sphere radius

    returns a ``yapcad.geom`` point
    """
    if lat == 90:
        return [0,0,rad,1]
    elif lat == -90:
        return [0,0,-rad,1]
    else:
        latr = (((lat+90)%180)-90)*pi2/360.0
        lonr = (lon%360)*pi2/360.0
    
        smallrad = math.cos(latr)*rad
        z = math.sin(latr)*rad
        x = math.cos(lonr)*smallrad
        y = math.sin(lonr)*smallrad
        return [x,y,z,1]

## icosohedron-specific function, generate intial geometry
def makeIcoPoints(center,radius):
    """
    Procedure to generate the verticies of an icosohedron with the specified
    ``center`` and ``radius``.
    """
    points = []
    normals = []
    p = sphere2cartesian(90,0,radius)
    n = scale3(p,1.0/mag(p))
    n[3] = 0.0
    points.append(p)
    normals.append(n)
    for i in range(10):
        sgn = 1
        if i%2 == 0:
            sgn =-1
        lat = math.atan(0.5)*360.0*sgn/pi2
        lon = i*36.0
        p = sphere2cartesian(lat,lon,radius)
        n = scale3(p,1.0/mag(p))
        n[3] = 0.0
        points.append(p)
        normals.append(n)

    p = sphere2cartesian(-90,0,radius)
    n = scale3(p,1.0/mag(p))
    n[3] = 0.0
    points.append(p)
    normals.append(n)
    return list(map( lambda x: add(x,center),points)),normals

# face indices for icosahedron
icaIndices = [ [1,11,3],[3,11,5],[5,11,7],[7,11,9],[9,11,1],
               [2,1,3],[2,3,4],[4,3,5],[4,5,6],[6,5,7],[6,7,8],[8,7,9],[8,9,10],[10,9,1],[10,1,2],
               [0,2,4],[0,4,6],[0,6,8],[0,8,10],[0,10,2] ]

vertexHash = {}
_vertexHash_owner = None
def addVertex(nv,nn,verts,normals):
    """
    Utility function that takes a vertex and associated normal and a
    list of corresponding vertices and normals, and returns an index
    that corresponds to the given vertex/normal.  If the vertex
    doesn't exist in the list, the lists are updated to include it and
    the corresponding normal.

    returns the index, and the (potentiall updated) lists
    """
    global vertexHash, _vertexHash_owner
    owner_id = id(verts)
    if _vertexHash_owner != owner_id or len(verts) == 0:
        vertexHash = {}
        _vertexHash_owner = owner_id

    found = False
    # Normalize tiny values to avoid "-0.00" != "0.00" hash key mismatch
    x = 0.0 if abs(nv[0]) < epsilon else nv[0]
    y = 0.0 if abs(nv[1]) < epsilon else nv[1]
    z = 0.0 if abs(nv[2]) < epsilon else nv[2]
    vkey = f"{x:.2f}{y:.2f}{z:.2f}"
    if vkey in vertexHash:
        found = True
        inds = vertexHash[vkey]
        valid_inds = []
        for i in inds:
            if i >= len(verts):
                continue
            if vclose(nv,verts[i]):
                return i,verts,normals
            valid_inds.append(i)
        vertexHash[vkey] = valid_inds
            
    verts.append(nv)
    normals.append(nn)
    i = len(verts)-1
    if found:
        vertexHash[vkey] += [ i ]
    else:
        vertexHash[vkey] = [ i ]
    return i,verts,normals

def subdivide(f,verts,normals,rad):
    """
    Given a face (a list of three vertex indices), a list of vertices,
    normals, and a radius, subdivide that face into four new faces and
    update the lists of vertices and normals accordingly.

    return the updated vertex and normal lists, and a list of the four
    new faces.
    """

    ind1 = f[0]
    ind2 = f[1]
    ind3 = f[2]
    v1 = verts[ind1]
    v2 = verts[ind2]
    v3 = verts[ind3]
    n1 = normals[ind1]
    n2 = normals[ind2]
    n3 = normals[ind3]
    va = add(v1,v2)
    vb = add(v2,v3)
    vc = add(v3,v1)
    ma = rad/mag(va)
    mb = rad/mag(vb)
    mc = rad/mag(vc)
    va = scale3(va,ma)
    vb = scale3(vb,mb)
    vc = scale3(vc,mc)
    
    na = add4(n1,n2)
    na = scale4(na,1.0/mag(na))
    nb = add4(n2,n3)
    nb = scale4(nb,1.0/mag(nb))
    nc = add4(n3,n1)
    nc = scale4(nc,1.0/mag(nc))

    inda,verts,normals = addVertex(va,na,verts,normals)
    indb,verts,normals = addVertex(vb,nb,verts,normals)
    indc,verts,normals = addVertex(vc,nc,verts,normals)

    f1 = [ind1,inda,indc]
    f2 = [inda,ind2,indb]
    f3 = [indb,ind3,indc]
    f4 = [inda,indb,indc]

    return verts,normals, [f1,f2,f3,f4]

# make the sphere, return a surface representation
def sphereSurface(diameter,center=point(0,0,0),depth=2):
    rad = diameter/2
    ## subdivision works only when center is origin, so we add
    ## any center offset after we subdivide
    verts,normals = makeIcoPoints(point(0,0,0),rad)
    faces = icaIndices

    for i in range(depth):
        ff = []
        for f in faces:
            verts, norms, newfaces = subdivide(f,verts,normals,rad)
            ff+=newfaces
        faces = ff
                      
    if not vclose(center,point(0,0,0)):
        verts = list(map(lambda x: add(x,center),verts))
    return ['surface',verts,normals,faces,[],[]]

# make sphere, return solid representation
def sphere(diameter,center=point(0,0,0),depth=2):
    call = f"yapcad.geom3d_util.sphere({diameter},center={center},depth={depth})"
    return solid( [ sphereSurface(diameter,center,depth)],
                  [],['procedure',call] )
                    

def rectangularPlane(length,width,center=point(0,0,0)):
    """ return a rectangular surface with the normals oriented in the
    positive z direction """
    c = center
    l = point(length,0,0)
    w = point(0,width,0)
    p0 = add(c,(scale3(add(l,w),-0.5)))
    p1 = add(p0,l)
    p2 = add(p1,w)
    p3 = add(p0,w)
    n = vect(0,0,1,0)
    
    surf = surface( [p0,p1,p2,p3],[n,n,n,n],
                    [[0,1,2],
                     [2,3,0]])

    return surf

# make a rectangular prism from six surfaces, return a solid
def prism(length,width,height,center=point(0,0,0)):
    """make a rectangular prism solid composed of six independent 
    faces"""
    call = f"yapcad.geom3d_util.prism({length},{width},{height},{center})"

    l2 = length/2
    w2 = width/2
    h2 = height/2
    
    topS = rectangularPlane(length,width,point(0,0,h2))
    bottomS = rotatesurface(topS,180,axis=point(1,0,0))
    frontS = rectangularPlane(length,height,point(0,0,w2))
    frontS = rotatesurface(frontS,90,axis=point(1,0,0))
    backS = rotatesurface(frontS,180)
    rightS = rectangularPlane(height,width,point(0,0,l2))
    rightS = rotatesurface(rightS,90,axis=point(0,1,0))
    leftS = rotatesurface(rightS,180)

    surfaces = [topS,bottomS,frontS,backS,rightS,leftS]
    if not vclose(center,[0,0,0,1]):
        surfaces = list(map(lambda x: translatesurface(x,center),surfaces))
    
    sol = solid(surfaces,
                [],
                ['procedure',call])
    return sol

def circleSurface(center,radius,angr=10,zup=True):
    """make a circular surface centered at ``center`` lying in the XY
    plane with normals pointing in the positive z direction if ``zup
    == True``, negative z otherwise"""

    if angr < 1 or angr > 45:
        raise ValueError('angular resolution must be between 1 and 45 degrees')
    
    samples = round(360.0/angr)
    angr = 360.0/samples
    basep=[center]
    
    for i in range(samples):
        theta = i*angr*pi2/360.0
        pp = [math.cos(theta)*radius,math.sin(theta)*radius,0.0,1.0]
        pp = add(pp,center)
        basep.append(pp)       

    
    basef=[]

    rng = range(1,len(basep))
    if not zup:
        rng = reversed(rng)

    ll = len(basep)-1
    for i in rng:
        face = []
        if zup:
            face = [0,i,1+i%ll]
        else:
            face = [0,1+i%ll,i]
            
        basef.append(face)

    z=-1
    if zup:
        z=1
    n = vect(0,0,z,0)
    basen= [ n ] * len(basep)

    return surface(basep,basen,basef)

def conic(baser,topr,height, center=point(0,0,0),angr=10):

    """Make a conic frustum splid, center is center of first 'base'
    circle, main axis aligns with positive z.  This function can be
    used to make a cone, a conic frustum, or a cylinder, depending on
    the parameters.

         ``baser`` is the radius of the base, must be greater than
         zero (epsilon).

         ``topr`` is the radius of the top, may be zero or
         positive. If ``0 <= topr < epsilon``, then the top
         is treated as a single point.

         ``height`` is distance from base to top, must be greater than
         epsilon.
    
         ``center`` is the location of the center of the base.

         ``angr`` is the requested angular resolution in degrees for
         sampling circles.  Actual angular resolution will be
         ``360/round(360/angr)``

    """
    call = f"yapcad.geom3d_util.conic({baser},{topr},{height},{center},{angr})"
    if baser < epsilon:
        raise ValueError('bad base radius for conic')
    base = arc(center,baser)

    toppoint = False
    if topr < 0:
        raise ValueError('bad top radius for conic')
    if topr < epsilon:
        toppoint = True

    if height < epsilon:
        raise ValueError('bad height in conic')

    baseS = circleSurface(center,baser,zup=False)
    baseV = baseS[1]
    ll = len(baseV)
        
    if not toppoint:
        topS = circleSurface(add(center,point(0,0,height)),
                             topr,zup=True)
        topV = topS[1]
        cylV = baseV[1:] + topV[1:]
        ll = ll-1
        baseN = []
        topN = []
        cylF = []
        for i in range(ll):
            p0 = cylV[(i-1)%ll]
            p1 = cylV[(i+1)%ll]
            p2 = cylV[ll+i]

            cylF.append([i,(i+1)%ll,ll+(i+1)%ll])
            cylF.append([i,ll+(i+1)%ll,ll+i])

            pp,n0 = tri2p0n([p0,p1,p2])
            
            baseN.append(n0)
            topN.append(n0)

        cylN = baseN+topN

        cylS = surface(cylV,cylN,cylF)

        return solid([baseS,cylS,topS],[],
                     ['procedure',call])
    else:
        topP = add(center,point(0,0,height))
        # Only use perimeter vertices (baseV[1:]), skip the center point
        conV = [ topP ] + baseV[1:]
        ll = len(conV)
        # Initialize all normals to a default value
        conN = [[0,0,1,0] for _ in range(ll)]
        conF = []

        # ll = apex(1) + perimeter vertices(36) = 37
        # Perimeter vertex indices: 1 to ll-1
        num_perimeter = ll - 1
        for i in range(1, ll):
            p0 = conV[0]  # apex
            # Wrap indices within perimeter range [1, ll-1]
            prev_idx = ((i - 2) % num_perimeter) + 1
            next_idx = (i % num_perimeter) + 1
            p1 = conV[prev_idx]
            p2 = conV[next_idx]

            try:
                pp, n0 = tri2p0n([p0, p1, p2])
            except ValueError:
                # Skip degenerate faces near the apex
                continue

            conF.append([0, i, next_idx])
            conN[i] = n0

        conS = surface(conV,conN,conF)

        return solid([baseS,conS],[],
                     ['procedure',call])

def makeRevolutionSurface(contour,zStart,zEnd,steps,arcSamples=36):
    """
    Take a countour (any function z->y mapped over the interval

     ``zStart`` and ``zEnd`` and produce the surface of revolution
     around the z axis.  Sample ``steps`` contours of the function,
    which in turn are turned into circles sampled `arcSamples`` times.
    """

    sV=[]
    sN=[]
    sF=[]
    zRange = zEnd-zStart
    zD = zRange/steps

    degStep = 360.0/arcSamples
    radStep = pi2/arcSamples

    # Pre-compute cos/sin values to avoid floating point errors at the seam
    # Explicitly ensure that index 0 uses exact values
    angle_cos = []
    angle_sin = []
    for i in range(arcSamples):
        if i == 0:
            angle_cos.append(1.0)
            angle_sin.append(0.0)
        else:
            angle = i * radStep
            angle_cos.append(math.cos(angle))
            angle_sin.append(math.sin(angle))

    # Check if we need pole caps
    r_start = contour(zStart)
    r_end = contour(zEnd)
    need_start_cap = r_start < epsilon * 10
    need_end_cap = r_end < epsilon * 10

    # Add pole vertices if needed
    start_pole_idx = None
    end_pole_idx = None
    if need_start_cap:
        pole_point = [0.0, 0.0, zStart, 1.0]
        pole_normal = [0.0, 0.0, -1.0, 0.0]  # Points downward for bottom pole
        start_pole_idx, sV, sN = addVertex(pole_point, pole_normal, sV, sN)

    if need_end_cap:
        pole_point = [0.0, 0.0, zEnd, 1.0]
        pole_normal = [0.0, 0.0, 1.0, 0.0]  # Points upward for top pole
        end_pole_idx, sV, sN = addVertex(pole_point, pole_normal, sV, sN)

    for i in range(steps):
        z = i*zD+zStart
        r0 = contour(z)
        r1 = contour(z+zD)

        # Handle pole caps
        if i == 0 and need_start_cap:
            # Create triangular faces from pole to first ring
            if r1 < epsilon:
                r1 = epsilon
            for j in range(arcSamples):
                a1_idx = j
                a2_idx = (j+1) % arcSamples

                pp1 = [angle_cos[a1_idx]*r1, angle_sin[a1_idx]*r1, z+zD, 1.0]
                pp2 = [angle_cos[a2_idx]*r1, angle_sin[a2_idx]*r1, z+zD, 1.0]

                try:
                    _, n = tri2p0n([sV[start_pole_idx], pp1, pp2])
                except ValueError:
                    continue

                k1, sV, sN = addVertex(pp1, n, sV, sN)
                k2, sV, sN = addVertex(pp2, n, sV, sN)
                sF.append([start_pole_idx, k1, k2])
            continue

        if i == steps - 1 and need_end_cap:
            # Create triangular faces from last ring to pole
            if r0 < epsilon:
                r0 = epsilon
            for j in range(arcSamples):
                a1_idx = j
                a2_idx = (j+1) % arcSamples

                p1 = [angle_cos[a1_idx]*r0, angle_sin[a1_idx]*r0, z, 1.0]
                p2 = [angle_cos[a2_idx]*r0, angle_sin[a2_idx]*r0, z, 1.0]

                try:
                    _, n = tri2p0n([p1, sV[end_pole_idx], p2])
                except ValueError:
                    continue

                k1, sV, sN = addVertex(p1, n, sV, sN)
                k2, sV, sN = addVertex(p2, n, sV, sN)
                sF.append([k1, end_pole_idx, k2])
            continue

        # Regular quad strips for non-pole sections
        if r0 < epsilon:
            r0 = epsilon
        if r1 < epsilon:
            r1 = epsilon

        for j in range(arcSamples):
            # Use pre-computed values with proper wrapping
            a0_idx = (j-1) % arcSamples
            a1_idx = j
            a2_idx = (j+1) % arcSamples

            p0 = [angle_cos[a0_idx]*r0, angle_sin[a0_idx]*r0, z, 1.0]
            p1 = [angle_cos[a1_idx]*r0, angle_sin[a1_idx]*r0, z, 1.0]
            p2 = [angle_cos[a2_idx]*r0, angle_sin[a2_idx]*r0, z, 1.0]

            pp1 = [angle_cos[a1_idx]*r1, angle_sin[a1_idx]*r1, z+zD, 1.0]
            pp2 = [angle_cos[a2_idx]*r1, angle_sin[a2_idx]*r1, z+zD, 1.0]

            try:
                p,n = tri2p0n([p0,p2,pp1])
            except ValueError:
                # Skip degenerate faces
                continue

            k1,sV,sN = addVertex(p1,n,sV,sN)
            k2,sV,sN = addVertex(p2,n,sV,sN)
            k3,sV,sN = addVertex(pp2,n,sV,sN)
            k4,sV,sN = addVertex(pp1,n,sV,sN)
            sF.append([k1,k2,k3])
            sF.append([k1,k3,k4])
        
    return surface(sV,sN,sF)

def contour(poly,distance,direction, samples,scalefunc= lambda x: (1,1,1)):
    """take a closed polygon and apply a scaling function defined on the
    interval 0,1 that returns a tuple of x,y,z scaling values. For
    each of ``samples`` number of samples, translate in ``direction``
    direction and scale the contour according to ``scalefunc()``,
    producing a surface."""
    if not ispolygon(poly):
        raise ValueError('invalid polygon passed to contour')
    if samples < 2:
        raise ValueError('number of samples must be 2 or greater')
    if not close(mag(direction),1.0):
        raise ValueError('bad direction vector')
    if distance <= epsilon:
        raise ValueError('bad distance passed to contour')

    raise NotImplemented("this function is a work in progress")

    p0 = poly
    p1 = []

    u = 0.0
    scle = scalefunc(u)
    sctx = Scale(scle[0],scle[1],scle[2])
    ply = list(map(lambda p: sctx.mul(p),poly))
    surf = poly2surface(ply)
    s1 = reversesurface(surf)

    u = 1.0
    scle = scalefunc(u)
    sctx = Scale(scle[0],scle[1],scle[2])
    ply2 = list(map(lambda p:  sctx.mul(p),poly))
    surf2 = poly2surface(ply2)
    s2 = translatesurface(surf2,scale4(direction,distance))
    
    vrts = surf[1]
    nrms = surf[2]
    facs = surf[3]
    bndr = surf[4]

    vrts1 = list(map(lambda i: vrts[i],bndr))
    for ii in range(1,samples):
        u = ii/samples
        
        stripF = []
        #vrts2 = list(map(lambda i:
        for i in range(len(bndr)):
            j0 = bndry1[(i-1)%len(bndry1)]
            j1 = bndry1[i]
            j2 = bndry1[(i+1)%len(bndry1)]
            j3 = j2+len(s2[1])
            j4 = j1+len(s2[1])
            p0 = stripV[j0]
            p1 = stripV[j2]
            p2 = stripV[j3]
            try:
                pp,n0 = tri2p0n([p0,p1,p2])
            except ValueError:
                # bad face, skip
                continue
            stripN[j1]=n0
            stripN[j4]=n0
            stripF.append([j1,j2,j3])
            stripF.append([j1,j3,j4])
            
            


def extrude(surf,distance,direction=vect(0,0,1,0)):

    """ Take a surface and extrude it in the specified direction to
    create a solid.  Return the solid. """
    call = f"yapcad.geom3d_util.extrude({surf},{distance},{direction})"

    if not issurface(surf):
        raise ValueError('invalid surface passed to extrude')

    if distance <= epsilon:
        raise ValueError('bad distance passed to extrude')

    s1 = translatesurface(surf,scale4(direction,distance))
    s2 = reversesurface(surf)

    loops = []
    if s2[4]:
        loops.append(list(s2[4]))
    loops.extend([list(loop) for loop in s2[5] if loop])

    stripV = s2[1] + s1[1]  # vertices for the edge strips
    stripN = [vect(0, 0, 1, 0)] * len(stripV)  # placeholder normals
    stripF: list[list[int]] = []
    offset = len(s2[1])

    def _project(idx):
        point3 = stripV[idx]
        return point3[0], point3[1]

    if surf[3]:
        try:
            tri = surf[3][0]
            basis = tri2p0n([surf[1][tri[0]], surf[1][tri[1]], surf[1][tri[2]]], basis=True)
            if basis:
                forward = basis[2]

                def _project(idx):  # type: ignore[redefinition]
                    vec = forward.mul(stripV[idx])
                    return vec[0], vec[1]
        except Exception:  # pragma: no cover
            pass

    def loop_area(loop):
        if len(loop) < 3:
            return 0.0
        area = 0.0
        for i in range(len(loop)):
            x0, y0 = _project(loop[i])
            x1, y1 = _project(loop[(i + 1) % len(loop)])
            area += x0 * y1 - x1 * y0
        return area / 2.0

    if loops:
        outer_area = loop_area(loops[0])
        if outer_area < 0:
            loops[0].reverse()
            outer_area = -outer_area
        outer_sign = 1 if outer_area >= 0 else -1
        for loop in loops[1:]:
            if loop_area(loop) * outer_sign > 0:
                loop.reverse()

    for bndry in loops:
        if len(bndry) < 2:
            continue
        for i in range(len(bndry)):
            j0 = bndry[(i - 1) % len(bndry)]
            j1 = bndry[i]
            j2 = bndry[(i + 1) % len(bndry)]
            j3 = j2 + offset
            j4 = j1 + offset
            p0 = stripV[j0]
            p1 = stripV[j2]
            p2 = stripV[j3]
            try:
                pp, n0 = tri2p0n([p0, p1, p2])
            except ValueError:
                continue
            stripN[j1] = n0
            stripN[j4] = n0
            stripF.append([j1, j2, j3])
            stripF.append([j1, j3, j4])

    #import pdb ; pdb.set_trace()
    strip = surface(stripV,stripN,stripF)

    return solid([s2,strip,s1],
                 [],
                 ['procedure',call])



def _loft_surface(lower_loop, upper_loop, invert=False):
    """Create a surface connecting two loops.

    Args:
        lower_loop: List of points forming the lower loop (must be open, not closed)
        upper_loop: List of points forming the upper loop (must be open, not closed)
        invert: If True, reverse the face winding order

    Returns:
        A yapCAD surface connecting the two loops with triangle strips
    """

    if not lower_loop or not upper_loop:
        raise ValueError('invalid loops passed to loft surface')
    lower = [point(p) for p in lower_loop]
    upper = [point(p) for p in upper_loop]
    if len(lower) != len(upper):
        raise ValueError('loop length mismatch in loft surface')

    vertices = lower + upper
    normals = [[0, 0, 1, 0] for _ in vertices]
    faces = []
    count = len(lower)

    for idx in range(count):
        j0 = idx
        j1 = (idx + 1) % count
        j2 = j1 + count
        j3 = idx + count

        tri1 = [j0, j1, j2]
        tri2 = [j0, j2, j3]
        if invert:
            tri1 = list(reversed(tri1))
            tri2 = list(reversed(tri2))

        try:
            _, normal = tri2p0n([vertices[tri1[0]],
                                 vertices[tri1[1]],
                                 vertices[tri1[2]]])
        except ValueError:
            continue

        for vid in {tri1[0], tri1[1], tri1[2], tri2[0], tri2[1], tri2[2]}:
            normals[vid] = normal

        faces.append(tri1)
        faces.append(tri2)

    return surface(vertices, normals, faces)


        

    

def _circle_loop(center_xy, radius, minang):
    arc_geom = [arc(point(center_xy[0], center_xy[1]), radius)]
    loop = geomlist2poly(arc_geom, minang=minang, minlen=0.0)
    if not loop:
        raise ValueError('failed to generate circle loop')
    return loop


def tube(outer_diameter, wall_thickness, length,
         center=None, *, base_point=None, minang=5.0, include_caps=True):
    """Create a cylindrical tube solid.

    ``base_point`` (or legacy ``center`` argument) identifies the base of the
    cylindrical wall, i.e. the plane where ``z == base_point[2]``.
    """

    if base_point is not None and center is not None:
        raise ValueError('specify only base_point (preferred) or center, not both')
    if base_point is None:
        base_point = center if center is not None else point(0, 0, 0)
    if len(base_point) < 3:
        raise ValueError('base_point must contain x, y, z components')
    base_point = point(base_point)

    if wall_thickness <= epsilon:
        raise ValueError('wall thickness must be positive')

    outer_radius = outer_diameter / 2.0
    inner_radius = outer_radius - wall_thickness
    if inner_radius <= epsilon:
        raise ValueError('wall thickness too large for tube')

    base_z = base_point[2]
    center_xy = (base_point[0], base_point[1])

    base_loop_xy = _circle_loop(center_xy, outer_radius, minang)
    inner_loop_xy = list(reversed(_circle_loop(center_xy, inner_radius, minang)))

    base_surface, _ = poly2surfaceXY(base_loop_xy, holepolys=[inner_loop_xy])
    base_surface = reversesurface(base_surface)
    base_surface = translatesurface(base_surface, point(0, 0, base_z))

    top_surface, _ = poly2surfaceXY(base_loop_xy, holepolys=[inner_loop_xy])
    top_surface = translatesurface(top_surface, point(0, 0, base_z + length))

    outer_base = [point(p[0], p[1], base_z, 1.0) for p in base_loop_xy[:-1]]
    inner_base = [point(p[0], p[1], base_z, 1.0) for p in inner_loop_xy[:-1]]
    outer_top = [point(p[0], p[1], base_z + length, 1.0) for p in base_loop_xy[:-1]]
    inner_top = [point(p[0], p[1], base_z + length, 1.0) for p in inner_loop_xy[:-1]]

    outer_side = _loft_surface(outer_base, outer_top)
    inner_side = _loft_surface(inner_top, inner_base, invert=True)

    call = f"yapcad.geom3d_util.tube({outer_diameter}, {wall_thickness}, {length}, base_point={base_point})"
    surfaces = [base_surface, outer_side, top_surface, inner_side]
    if not include_caps:
        surfaces = [outer_side, inner_side]
    return solid(surfaces,
                 [],
                 ['procedure', call])


def conic_tube(bottom_outer_diameter, top_outer_diameter, wall_thickness,
               length, center=None, *, base_point=None, minang=5.0, include_caps=True):
    """Create a conic tube with varying outer diameter.

    ``base_point`` (or ``center`` legacy argument) marks the axial base of the
    frustum (the larger-diameter end when stacked)."""

    if base_point is not None and center is not None:
        raise ValueError('specify only base_point (preferred) or center, not both')
    if base_point is None:
        base_point = center if center is not None else point(0, 0, 0)
    base_point = point(base_point)

    if wall_thickness <= epsilon:
        raise ValueError('wall thickness must be positive')

    r0_outer = bottom_outer_diameter / 2.0
    r1_outer = top_outer_diameter / 2.0
    r0_inner = r0_outer - wall_thickness
    r1_inner = r1_outer - wall_thickness
    if r0_inner <= epsilon or r1_inner <= epsilon:
        raise ValueError('wall thickness too large for conic tube')

    base_z = base_point[2]
    center_xy = (base_point[0], base_point[1])

    base_outer_loop = _circle_loop(center_xy, r0_outer, minang)
    base_inner_loop = list(reversed(_circle_loop(center_xy, r0_inner, minang)))

    top_outer_loop = _circle_loop(center_xy, r1_outer, minang)
    top_inner_loop = list(reversed(_circle_loop(center_xy, r1_inner, minang)))

    base_surface, _ = poly2surfaceXY(base_outer_loop, holepolys=[base_inner_loop])
    base_surface = reversesurface(base_surface)
    base_surface = translatesurface(base_surface, point(0, 0, base_z))

    top_surface, _ = poly2surfaceXY(top_outer_loop, holepolys=[top_inner_loop])
    top_surface = translatesurface(top_surface, point(0, 0, base_z + length))

    outer_base = [point(p[0], p[1], base_z, 1.0) for p in base_outer_loop[:-1]]
    outer_top = [point(p[0], p[1], base_z + length, 1.0) for p in top_outer_loop[:-1]]
    inner_base = [point(p[0], p[1], base_z, 1.0) for p in base_inner_loop[:-1]]
    inner_top = [point(p[0], p[1], base_z + length, 1.0) for p in top_inner_loop[:-1]]

    outer_side = _loft_surface(outer_base, outer_top)
    inner_side = _loft_surface(inner_top, inner_base, invert=True)

    call = ("yapcad.geom3d_util.conic_tube("
            f"{bottom_outer_diameter}, {top_outer_diameter}, {wall_thickness}, {length}, base_point={base_point})")
    surfaces = [base_surface, outer_side, top_surface, inner_side]
    if not include_caps:
        surfaces = [outer_side, inner_side]
    return solid(surfaces,
                 [],
                 ['procedure', call])


def spherical_shell(outer_diameter, wall_thickness,
                    solid_angle=4 * math.pi, center=point(0, 0, 0),
                    *, minang=5.0, steps=24):
    """Create a spherical shell or cap defined by a solid angle."""

    if wall_thickness <= epsilon:
        raise ValueError('wall thickness must be positive')

    outer_radius = outer_diameter / 2.0
    inner_radius = outer_radius - wall_thickness
    if inner_radius <= epsilon:
        raise ValueError('wall thickness too large for spherical shell')

    solid_angle = max(min(solid_angle, 4 * math.pi), epsilon)
    cos_theta = 1.0 - solid_angle / (2.0 * math.pi)
    cos_theta = max(-1.0, min(1.0, cos_theta))
    theta = math.acos(cos_theta)

    arc_samples = max(12, int(round(360.0 / minang)))
    lat_steps = max(4, int(math.ceil(theta / math.radians(minang))))

    cz = center[2]

    def _sphere_contour(radius):
        def contour(z):
            dz = z - cz
            dz = max(min(dz, radius), -radius)
            return math.sqrt(max(radius * radius - dz * dz, 0.0))
        return contour

    z_start_outer = cz + outer_radius * math.cos(theta)
    z_start_inner = cz + inner_radius * math.cos(theta)
    z_end_outer = cz + outer_radius
    z_end_inner = cz + inner_radius

    outer_surface = makeRevolutionSurface(_sphere_contour(outer_radius),
                                          z_start_outer, z_end_outer,
                                          max(steps, lat_steps),
                                          arcSamples=arc_samples)
    outer_surface = translatesurface(outer_surface, point(center[0], center[1], 0))

    inner_surface = makeRevolutionSurface(_sphere_contour(inner_radius),
                                          z_start_inner, z_end_inner,
                                          max(steps, lat_steps),
                                          arcSamples=arc_samples)
    inner_surface = translatesurface(inner_surface, point(center[0], center[1], 0))
    inner_surface = reversesurface(inner_surface)

    surfaces = [outer_surface, inner_surface]

    if theta < math.pi - epsilon:
        r_outer_ring = outer_radius * math.sin(theta)
        r_inner_ring = inner_radius * math.sin(theta)

        base_outer = [point(center[0] + p[0], center[1] + p[1], cz + outer_radius * math.cos(theta), 1.0)
                      for p in _circle_loop((0, 0), r_outer_ring, minang)[:-1]]
        base_inner = [point(center[0] + p[0], center[1] + p[1], cz + inner_radius * math.cos(theta), 1.0)
#                      for p in list(reversed(_circle_loop((0, 0), r_inner_ring, minang)))[:-1]]
                      for p in _circle_loop((0, 0), r_inner_ring, minang)[:-1]]

        conic_surface = _loft_surface(base_outer, base_inner, invert=False)
        surfaces.append(conic_surface)

    call = ("yapcad.geom3d_util.spherical_shell("
            f"{outer_diameter}, {wall_thickness}, {solid_angle}, center={center})")
    return solid(surfaces,
                 [],
                 ['procedure', call])


def stack_solids(solids, *, axis='z', start=0.0, gap=0.0, align='center'):
    """Return translated copies of ``solids`` stacked along an axis."""

    if not solids:
        return []

    axis = axis.lower()
    if axis not in ('x', 'y', 'z'):
        raise ValueError('axis must be one of x, y, or z')

    axis_idx = {'x': 0, 'y': 1, 'z': 2}[axis]
    other_idx = [i for i in range(3) if i != axis_idx]

    placed = []
    cursor = start
    reference = None
    pending_gap = 0.0

    for entry in solids:
        if isinstance(entry, str):
            directive = entry.strip().lower()
            if directive.startswith('space:'):
                try:
                    value = float(directive.split(':', 1)[1])
                except ValueError as exc:
                    raise ValueError(f'bad spacing directive {entry}') from exc
                pending_gap += value
                continue
            raise ValueError(f'unsupported directive {entry!r} in stack_solids')

        solid_obj = entry
        bbox = solidbbox(solid_obj)
        length = bbox[1][axis_idx] - bbox[0][axis_idx]
        if length < epsilon:
            raise ValueError('solid has zero length along stacking axis')

        cursor += pending_gap
        pending_gap = 0.0

        translation = [0.0, 0.0, 0.0]
        translation[axis_idx] = cursor - bbox[0][axis_idx]

        if reference is None:
            if align == 'center':
                reference = [
                    (bbox[0][idx] + bbox[1][idx]) / 2.0 for idx in other_idx
                ]
            elif align == 'min':
                reference = [bbox[0][idx] for idx in other_idx]
            elif align == 'max':
                reference = [bbox[1][idx] for idx in other_idx]
            else:
                raise ValueError('align must be center, min, or max')

        if align == 'center':
            for ref_val, idx in zip(reference, other_idx):
                translation[idx] = ref_val - (bbox[0][idx] + bbox[1][idx]) / 2.0
        elif align == 'min':
            for ref_val, idx in zip(reference, other_idx):
                translation[idx] = ref_val - bbox[0][idx]
        elif align == 'max':
            for ref_val, idx in zip(reference, other_idx):
                translation[idx] = ref_val - bbox[1][idx]

        placed.append(translatesolid(solid_obj, vect(translation[0], translation[1], translation[2], 0)))
        cursor += length + gap

    return placed



"""Demonstration of yapCAD solid boolean operations.

This script can visualise the result via OpenGL or export the mesh to STL/STEP.
"""

import argparse
from pathlib import Path

from yapcad.geom import point
from yapcad.geom3d import solid_boolean, translatesolid
from yapcad.geom3d_util import prism, sphere, conic, tube
from yapcad.io.stl import write_stl
from yapcad.io.step import write_step


def _build_solids(kind: str):
    if kind == 'cube':
        a = prism(2, 2, 2)
        b = translatesolid(prism(2, 2, 2), point(0.75, 0.0, 0.0))
    elif kind == 'sphere':
        a = sphere(2.0)
        b = translatesolid(sphere(2.0), point(1.0, 0.0, 0.0))
    elif kind == 'box_hole':
        a = prism(2, 2, 2)
        b = conic(0.6, 0.6, 3.0, center=point(0.0, 0.0, -1.5))
    elif kind == 'tube_hole':
        a = tube(outer_diameter=3.0, wall_thickness=0.5, length=4.0,
                 base_point=point(0.0, 0.0, -2.0))
        b = conic(0.6, 0.6, 4.2, center=point(1.2, 0.0, -2.1))
    else:
        raise ValueError(f'unsupported shape kind: {kind!r}')
    return a, b


def _render_opengl(solids, result):
    from yapcad.pyglet_drawable import pygletDraw

    viewer = pygletDraw()
    viewer.make_object('obj1', lighting=True, linecolor='white',
                       material='pearl',
                       position=point(-2.5, 2.5, 0))

    viewer.make_object('obj2', lighting=True, linecolor='white',
                       material='obsidian',
                       position=point(2.5, 2.5, 0))

    viewer.make_object('rslt', lighting=True, linecolor='white',
                       material='gold',
                       position=point(0, 0, 0))

    viewer.draw_solid(solids[0], name='obj1')
    viewer.draw_solid(solids[1], name='obj2')
    viewer.draw_solid(result, name='rslt')
    viewer.display()


def _export_stl(result, output):
    path = Path(output)
    if path.suffix.lower() != '.stl':
        path = path.with_suffix('.stl')
    write_stl(result, path)
    print(f'Wrote STL to {path}')


def _export_step(result, output):
    path = Path(output)
    if path.suffix.lower() not in {'.step', '.stp'}:
        path = path.with_suffix('.step')
    write_step(result, path)
    print(f'Wrote STEP to {path}')


def main():
    parser = argparse.ArgumentParser(description='Solid boolean demonstration.')
    parser.add_argument('--mode', choices=['gl', 'stl', 'step'], default='gl',
                        help='visualise with OpenGL or export as STL/STEP')
    parser.add_argument('--operation', choices=['union', 'intersection', 'difference'],
                        default='union', help='boolean operation to apply')
    parser.add_argument('--shapes', choices=['cube', 'sphere', 'box_hole', 'tube_hole'],
                        default='cube',
                        help='primitive pair to use')
    parser.add_argument('--output', default='boolean_result',
                        help='output basename for STL/STEP export')
    parser.add_argument('--stitch', action='store_true',
                        help='experimentally stitch open edges in the result')
    parser.add_argument('--engine', default=None,
                        help='boolean engine to use (default: native)')
    args = parser.parse_args()

    solids = _build_solids(args.shapes)
    result = solid_boolean(solids[0], solids[1], args.operation,
                           stitch=args.stitch, engine=args.engine)

    if args.mode == 'gl':
        _render_opengl(solids, result)
    elif args.mode == 'stl':
        _export_stl(result, args.output)
    else:
        _export_step(result, args.output)


if __name__ == '__main__':
    main()



## Derived Geometry clases for yapCAD
from yapcad.geom import *
from yapcad.geometry import *

## Copyright (c) 2020 Richard W. DeVaul
## Copyright (c) 2020 yapCAD contributors
## All rights reserved

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Derived Geometry Classes
========================

These subclasses of Geometry compute a figure based on the contents
of their elements list, rather than simply providing a wrapper around
the contents of their elements.  

The Polygon class is an intepolating closed figure class that will
interpolate lines between each point or open figure element in its
element list.  It will also treat circles as rounded convex conrners
of the specified radius.  

By treating circles as rounded corners, polygon instances can be
"grown" to create new figures that are a fixed distance larger, which
is very useful for offsetting drill holes from the edge of a boundary,
etc.  """

class Polygon(Geometry):
    """Multi-element closed figure derived geometry class"""

    def __init__(self,a=None):
        super().__init__()
        self._setDerived(True)
        self._setClosed(True)
        self.__lengths=0
        self.__outline=[]
        
        if isinstance(a,Polygon):
            self._setElem(deepcopy(a.elem))
            self._setUpdate(True)
        elif isgeomlist(a):
            for i in a:
                if ispoint(i) or isarc(i):
                    self.elem.append(i)
                elif isline(i):
                    pass
                elif ispoly(i):
                    self._setElem(self.elem + i)
                else:
                    raise ValueError("bad argument to Polygon constructor")

        elif a != None:
            raise ValueError("bad argument to Polygon constructor")

        elm = self.elem

        l = len(elm)
        if l > 2:
            # if first and last element are identicial points or circles,
            # remove them
            if ((ispoint(elm[0]) and ispoint(elm[-1]) and 
                 dist(elm[0],elm[-1]) < epsilon) or 
                (iscircle(elm[0]) and iscircle(elm[-1]) and 
                 abs(elm[0][1][0] - elm[-1][1][0]) < epsilon)):
                self.elem.pop()
        
    def __repr__(self):
        return 'Polygon({})'.format(vstr(self.elem))

    ## compute barycentric center of figure by equally weighting the
    ## points.  If last and first points are the same, ignore the
    ## last point.
    def _calcCenter(self):
        l = len(self.elem)
        if l == 0:
            return None
        elif l == 1:
            return center(self.elem[0]) # center is sole point
        else:
            
            if dist(center(self.elem[0]),
                    center(self.elem[-1])) < epsilon:
                l -= 1

            
            p = center(self.elem[0])
            for i in range(1,l):
                p = add(center(self.elem[i]),p)

            return scale3(p,1/l)

    
    def _updateInternals(self):
        if self.update:
            self._setUpdate(False)
            self._makeoutline()
            self._setCenter(self._calcCenter())
            self._setBbox(bbox(self.__outline))

    def addPoint(self,element):
        if ispoint(element):
            self._setUpdate(True)  # flag that we need to recalculate stuff
            self.elem.append(deepcopy(element))
        else:
            raise ValueError('attempt to add a non point to Polyline')

    ## set a point to a specified value
    def setPoint(self,i,p):
        if not ispoint(p):
            raise ValueError('bad point passed to Polyline.setPoint(): '.format(p))
        elem = self.elem
        if i < len(elem):
            elem[i]=point(p)
        elif i == len(elem):
            elem.append(p)
        else:
            raise ValueError('index out of range in PolylinesetPoint(): '.format(i))
        self._setElem(elem)
        self._setUpdate(True)

    ## add another drawing element
    def addLine(self,element):
        if isline(element):
            self._setUpdate(True)  # flag that we need to recalculate stuff
            self.elem.append(deepcopy(element))
        else:
            raise ValueError('attempt to add a non point, line or arc to poly')
    
    ## add another drawing element
    def addArc(self,element):
        if isarc(element):
            self._setUpdate(True)  # flag that we need to recalculate stuff
            self.elem.append(deepcopy(element))
        else:
            raise ValueError('attempt to add a non point, line or arc to poly')
                
    def remove(self,element):
        self._setUpdate(True)
        self.elem.remove(element)
                
    ## function to take the elements in the elem[] list and
    ## construct the full outline.  For example, consider a list of
    ## three points in the elem[] list -- this would result in the
    ## construction of three lines that results in a closed curve.
    ## Similarly, three circles in the elem list would result in a
    ## "rounded triangle" composed of the circle-tangent lines joined
    ## by three arcs.  Elements that are explicitly specified lines or
    ## non-circular arcs are joined to adjacent elements by lines

    def _makeoutline(self):
        # import pdb; pdb.set_trace()
        def _calclength():
            l = 0
            ll = []
            length = 0
            for o in self.__outline:
                if isarc(o):
                    length=arclength(o)
                elif isline(o): # line
                    length=linelength(o)
                else:
                    #print("self.__outline: ",vstr(self.__outline))
                    #print("o: ",vstr(o))
                    raise ValueError("bad element in outline list for _calclength()")
                l += length
                ll.append(length)
            self._setLength(l)
            self.__lengths=ll

        def _handleCircle(e0,e1,e2):
            ## get the two tangent lines from circle e1 to the circle
            ## e2
            ll = circleCircleTangentsXY(e1,e2)
            l=[]
            x0=center(e0)
            x1=center(e1)
            x2=center(e2)
            #print("x0: ",vstr(x0)," x1: ",vstr(x1)," x2: ",vstr(x2))
            #print("ll: ",vstr(ll))
            v1= sub(x1,x0)
            v2= sub(x2,x1)
            r0 = cross(v1,v2)
            x3 = linecenter(ll[0])
            x4 = linecenter(ll[1])
            #print("x3: ",vstr(x3),"x4: ",vstr(x4))
            #print("v1: ",vstr(v1),"v2: ",vstr(v2))
            v3 = sub(x3,x2)
            v4 = sub(x4,x2)
            r1 = cross(v2,v3)
            r2 = cross(v2,v4)
            #print("v3: ",vstr(v3)," v4: ",vstr(v4))
            #print("r0: ", r0[2], " r1: ",r1[2]," r2: ",r2[2])

            if r0[2] >= 0:
                #print("left-hand turn")
                if r1[2] >= 0:
                    l = ll[1]
                else:
                    l = ll[0]
            else:
                #print("right hand turn")
                if r2[2] >= 0:
                    l = ll[0]
                else:
                    l = ll[1]
                        
            self.__outline.append(line(l))
            
        def _fromPointAdd(e0,p1,e2):
            if ispoint(e2): #r1 is a point -- simplest case
                #ll = dist(p1,e2)
                self.__outline.append(line(p1,e2))
            elif isline(e2):
                self.__outline.append(line(p1,e2[0]))
                self.__outline.append(line(e2))
            elif isarc(e2) and not iscircle(e2):
                p = samplearc(e2,1*epsilon)
                self.__outline.append(line(p1,p))
                self.__outline.append(arc(e2))
            elif iscircle(e2): 
                _handleCircle(e0,arc(p1,1*epsilon),e2)
                self.__outline.append(arc(e2))
            else:
                raise ValueError('bad object in element list')
            
        self.__outline=[]
        self._setLength(0.0)
        self.__lengths=[]

        if len(self.elem) < 2:
            ## if one element, the outline is the element
            if len(self.elem) == 1:
                self.__outline=deepcopy(self.elem)
                _calclength()
            ## if there are fewer than one elem, there is nothing to do
            return
        ## we have work to do.  Construct all lines and circle-tangent
        ## lines first, inserting explicit line and arc elements.
        ## Then go back and convert circles to arcs.  Finally compute
        ## total length and length list for use in sample() function.

        
        for i in range(len(self.elem)):
            if i == 0:
                e0 = self.elem[-1]
            else:
                e0 = self.elem[i-1]
            e1 = self.elem[i]
            if i == len(self.elem)-1: #last item
                e2 = self.elem[0]
            else:
                e2 = self.elem[i+1]
            ## work through element types
            if ispoint(e1): #e1 is a point
                _fromPointAdd(e0,e1,e2)
            elif isline(e1):
                _fromPointAdd(e0,e1[0],e2)
            elif isarc(e1) and not iscircle(e1):
                p1 = samplearc(e1,1.0)
                _fromPointAdd(e0,p1,e2)
            elif iscircle(e1):
                c2 = []
                if ispoint(e2):
                    c2 = arc(e2,1*epsilon)
                if iscircle(e2):
                    c2 = e2
                else:
                    p = sample(e2,0.0)
                    c2 = arc(p,1*epsilon)
                _handleCircle(e0,e1,c2)
                
                if not ispoint(e2):
                    self.__outline.append(deepcopy(e2))

        ## OK, now go back through and replace full circles with arcs
        ## and catch any intersecting lines due to non-convex
        ## curvature
        for i in range(1,len(self.__outline)):
            e0 = self.__outline[i-1]
            e1 = self.__outline[i]
            if i == len(self.__outline)-1: #last item
                e2 = self.__outline[0]
            else:
                e2 = self.__outline[i+1]
            if isline(e1) and isline(e2):
                pi = lineLineIntersectXY(e1,e2,inside=True)
                if pi == False:
                    # print("not expected: adjacent lines don't intersect. We will fix that")
                    pi = lineLineIntersectXY(e1,e2,inside=False)
                    if pi == False:
                        # raise ValueError('parallel adjacent lines -- no clue')
                        print("odd parallel? adjacent line condition")
                        continue
                    else:
                        self.__outline[i] = line(e1[0],pi)
                        self.__outline[(i+1)%len(self.__outline)] = line(pi,e2[1])
            if iscircle(e1):
                if not isline(e0) or not isline(e2):
                    raise ValueError('circle not bracketed by lines')
                pp0 = lineArcIntersectXY(e0,e1,False)
                pp1 = lineArcIntersectXY(e2,e1,False)
                # these should be tangent lines, so exactly one
                # intersection each.  sometimes we get two that are
                # just a bit more than epsilon apart, which is OK.
                if len(pp0) < 1 or len(pp1) < 1:
                    raise ValueError('bad line-circle intersection in poly outline calculation')
                # we are assuming that elements are ordered in a
                # counter-clockwise fashion
                p0=sub(pp0[0],e1[0])
                p1=sub(pp1[0],e1[0])
                start = (atan2(p0[1],p0[0]) % pi2) * 360.0/pi2
                end = (atan2(p1[1],p1[0]) % pi2) * 360.0/pi2
                newarc = arc(e1[0],e1[1][0],start,end)
                self.__outline[i]=newarc

        self.__outline = cullZeroLength(self.__outline)
                
        _calclength()

    # sample function uses cached length and lengths[], otherwise
    # behavior is identical to superclass
    def sample(self,u):
        if self.update:
            self._updateInternals()
        if len(self.__outline) == 0:
            raise ValueError('no geometry to sample, empty poly')
        dist = (u % 1.0) * self.length
        d = 0
        for i in range(len(self.__outline)):
            l=self.__lengths[i]
            if dist <= d+l:
                u = 1.0 - (d+l-dist)/l
                e = self.__outline[i]
                return sample(e,u)
            else:
                d=d+l
                
    @property
    def geom(self):
        if self.update:
            self._updateInternals()
        return deepcopy(self.__outline)
                

    def grow(self,r):
        if close(r,0.0):
            return
        elif r < 0:
            raise ValueError('negative growth values not valid')
        def _growgeom(elem):
            ne = []
            for e in elem:
                if ispoint(e):
                    ne.append(arc(e,r))
                elif isline(e):
                    ne.append(arc(e[0],r))
                    ne.append(arc(e[1],r))
                elif isarc(e):
                    a = arc(e)
                    a[1][0] += r
                    ne.append(a)
                elif ispoly(e):
                    for p in e:
                        ne.append(arc(p,r))
                elif isgeomlist(e):
                    ne += _growgeom(e)
            return ne
        self._setElem(_growgeom(self.elem))
        self._setUpdate(True)

        
    def shrink(self,r):
        raise NotImplementedError('shrink operation not yet implemented')
        return False

    
## Utility functions


## function to process a geometry list and remove zero-length
## elements, including points, zero-length lines, and zero-length
## arcs.  Will recursively process geometry lists.

def cullZeroLength(geom):
    gl = []
    for g in geom:
        if ispoint(g):
            pass
        elif (isline(g) or isarc(g) or ispoly(g)) \
             and length(g) > epsilon:
            gl.append(deepcopy(g))
        elif isgeomlist(g) and length(g) > epsilon:
            gl.append(cullZeroLength(g))
        else: # zero-length or unknown element
            print("zero length element ",vstr(g)," culled")
    return gl



## make a Polygon instance containing a circle as two half-arcs

# def Circle(center=point(0,0,0),radius=1.0):
#     poly = Polygon()
#     poly.addArc(arc(center,radius,0.0,180.0))
#     poly.addArc(arc(center,radius,180.0,360.0))
#     return poly

def Circle(center=point(0,0,0),radius=1.0):
    return Geometry(arc(center,radius,0,360))

## make a rectangle with the specified width and height

def Rect(width,height,center=point(0,0,0)):
    w=width/2.0
    h=height/2.0
    p0=add(point(-w,h),center)
    p1=add(point(-w,-h),center)
    p2=add(point(w,-h),center)
    p3=add(point(w,h),center)
    poly = Polygon()
    poly.addPoint(p0)
    poly.addPoint(p1)
    poly.addPoint(p2)
    poly.addPoint(p3)
    return poly

## make rounded rectangle with specified width, height, and chamfer,
## centered at the origin by default

def RoundRect(width,height,chamf,center=point(0,0,0)):
    cr=chamf/2.0
    wid=width-chamf
    hei=height-chamf
    w=wid/2.0
    h=hei/2.0
    p0=add(point(-w,h),center)
    p1=add(point(-w,-h),center)
    p2=add(point(w,-h),center)
    p3=add(point(w,h),center)
    c0=arc(p0,cr)
    c1=arc(p1,cr)
    c2=arc(p2,cr)
    c3=arc(p3,cr)
    poly = Polygon()
    poly.addArc(c0)
    poly.addArc(c1)
    poly.addArc(c2)
    poly.addArc(c3)
    return poly



import pytest
import os
import random
from yapcad.geom import *
from yapcad.octtree import *

# Control flag for visual tests - set via environment variable or directly
VISUALTEST = os.environ.get('VISUALTEST', 'false').lower() in ('true', '1', 'yes')
if VISUALTEST:
    from yapcad.pyglet_drawable import pygletDraw

def three2two(x):
    """take a 3D bounding box with non-zero z coordinates, return
    a 2D bounding box with zero z coordinates"""
    a = vect(x[0])
    b = vect(x[1])
    a[2] = 0
    b[2] = 0
    return [a,b]

def randomPoints(bbox,numpoints):
    """Given a 3D bounding box and a number of points to generate, 
    return a list of uniformly generated random points within the 
    bounding box"""
    
    points = []
    minx = bbox[0][0]
    maxx = bbox[1][0]
    miny = bbox[0][1]
    maxy = bbox[1][1]
    minz = bbox[0][2]
    maxz = bbox[1][2]
    rangex = maxx-minx
    rangey = maxy-miny
    rangez = maxz-minz
    for i in range(numpoints):
        points.append(point(random.random()*rangex+minx,
                            random.random()*rangey+miny,
                            random.random()*rangez+minz))
    return points


class TestOctree:
    """ Test utility functions """

    ## create non-overlappig bounding boxes
    b1 = [point(-10,-5,-5),point(-5,5,5)]
    b2 = [point(-2.5,-2.5,-2.5),point(2.5,2.5,2.5)]
    b3 = [point(5,-5,-5),point(10,5,5)]
    
    b4 = [point(-2.5,-10,-5),point(2.5,-5,5)]
    b5 = [point(-2.5,5,-5),point(2.5,10,5)]

    b6 = [point(-3.5,-3.5,-10),point(3.5,3.5,-5)]
    b7 = [point(-3.5,-3.5,5),point(3.5,3.5,10)]

    # a box that contains b2
    b10=[point(-3,-3,-3),point(3,3,3)]

    # crossing boxes that penetrate b2 and others

    b20= [point(-15,-1,-1),point(15,1,1)]
    b21= [point(-1,-15,-1),point(1,15,1)]
    b22= [point(-1,-1,-15),point(1,1,15)]

    # all of the boxes
    
    glist = [b1,b2,b3,b4,b5,b6,b7,
             b10,
             b20,b21,b22]

    # no 3D overlap list

    gl_no_over = [b1,b2,b3,b4,b5,b6,b7]

    # 3D crossing list

    gl_cross = [b10,
                b20,b21,b22]

    # no 2D overlap list
    gl2D_no_over = [b1,b2,b3,b4,b5]
    gl2D_no_over = list(map(three2two,gl2D_no_over))

    # 2D crossing list
    gl2D_cross = [b10, b20, b21]
    gl2D_cross = list(map(three2two,gl2D_cross))

    @pytest.mark.visual
    def test_boxoverlap2(self):
        """ do bounding box overlap check testing """
        if not VISUALTEST:
            pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

        dd = pygletDraw()
        dd.linecolor = 'white'
        dd.cameradist = 35
    
        dd.draw_bbox(self.b1,dim3=True)
        dd.draw_bbox(self.b2,dim3=True)
        dd.draw_bbox(self.b3,dim3=True)
        dd.draw_bbox(self.b4,dim3=True)
        dd.draw_bbox(self.b5,dim3=True)
        dd.draw_bbox(self.b6,dim3=True)
        dd.draw_bbox(self.b7,dim3=True)

        dd.linecolor = 'red'
        dd.draw_bbox(self.b10,dim3=True)
        dd.linecolor = 'yellow'
        dd.draw_bbox(self.b20,dim3=True)
        dd.draw_bbox(self.b21,dim3=True)
        dd.draw_bbox(self.b22,dim3=True)
        
        dd.display()
    
        assert not boxoverlap2(self.b1,self.b2)
        assert not boxoverlap2(self.b1,self.b2,dim3=False)
        assert not boxoverlap2(self.b2,self.b3)
        assert not boxoverlap2(self.b2,self.b3,dim3=False)
        assert not boxoverlap2(self.b3,self.b4)
        assert not boxoverlap2(self.b3,self.b4,dim3=False)
        assert not boxoverlap2(self.b4,self.b2)
        assert not boxoverlap2(self.b4,self.b2,dim3=False)
        assert not boxoverlap2(self.b5,self.b2)
        assert not boxoverlap2(self.b5,self.b2,dim3=False)
    
        assert not boxoverlap2(self.b6,self.b2)
        # 2D box-in-box condition
        assert boxoverlap2(self.b6,self.b2,dim3=False)

        assert not boxoverlap2(self.b7,self.b2)
        # 2D box-in-box condition
        assert boxoverlap2(self.b7,self.b2,dim3=False)
        assert boxoverlap2(self.b2,self.b7,dim3=False)

        # 3D box-in-box condition
        assert boxoverlap2(self.b10,self.b2)
        assert boxoverlap2(self.b20,self.b2)
        assert boxoverlap2(self.b21,self.b2)
        assert boxoverlap2(self.b22,self.b2)

        bigbox = bbox(self.glist)
        # make sure computed geomlist bbox is correct, and assert that
        # all of the subboxes will be inside
        assert bigbox == [[-15,-15,-15,1],[15,15,15,1]]
        assert len(list(filter(lambda x: not boxoverlap2(bigbox,x),
                               self.glist))) == 0

    def test_bbox2oct(self):

        # find middle of box
        def boxmid(box):
            p0 = box[0]
            p1 = box[1]
            return scale3(add(p0,p1),0.5)

        bigbox = bbox(self.glist)
        center = boxmid(bigbox)

        inds = []
        for box in self.gl_no_over:
            inds.append(sorted(bbox2oct(box,bigbox,center)))

        # this is the correct list of octants for each test box
        assert inds[0] == [1,2,5,6]
        assert inds[1] == [0,1,2,3,4,5,6,7]
        assert inds[2] == [0,3,4,7]
        assert inds[3] == [2,3,6,7]
        assert inds[4] == [0,1,4,5]
        assert inds[5] == [0,1,2,3]
        assert inds[6] == [4,5,6,7]

        print("inds: ",inds)

    @pytest.mark.visual
    def test_quadtree(self):
        if not VISUALTEST:
            pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

        dd = pygletDraw()
        dd.cameradist = 35

        # make a quadtree
        bar = NTree(4)

        for l in self.gl2D_no_over:
            bar.addElement(l)

        bar.updateTree()
        dd.linecolor = 'red'
        for b in self.gl2D_no_over:
            dd.draw_bbox(b,dim3=False)
        dd.linecolor = 'white'
        dd.draw(self.gl2D_no_over)
        dd.display()

        # get full contents of tree
        bigbox = bbox(self.gl2D_no_over)
        contents = bar.getElements(bigbox)
        print("full contents: ",vstr(contents))
        assert len(contents) == len(self.gl2D_no_over)
        assert len(list(filter(lambda x: not x in self.gl2D_no_over,
                               contents))) == 0

        # retrieve the elements that overlap with self.b10

        contents = bar.getElements(self.b10)
        print("selected contents: ",vstr(contents))

        assert len(contents) == 1
        cexpect = [three2two(self.b2)]
        for c in contents:
            assert c in cexpect

        # retrieve the elements that overlap with self.b20

        contents = bar.getElements(self.b20)
        print("selected contents: ",vstr(contents))

        assert len(contents) == 3
        cexpect = list(map(three2two,[self.b1, self.b2, self.b3 ]))
        for c in contents:
            assert c in cexpect

               
        bar2 = NTree(4,mindim=4.0)
        # make 100 lines in space, keep track of which fall inside a
        # sub-region, then try to retrieve those with our tree,

        bigbox = [[-15,-15,0,1],[15,15,0,1]]
        deltabox = [[-10,-10,0,1],[10,10,0,1]]
        subbox = [[10,-10,0,1],[20,0,0,1]]
        glist=[]
        sublist=[]
        
        dim = 200
        plist1 = randomPoints(bigbox,dim)
        plist2 = randomPoints(deltabox,dim)
        
        for i in range(dim):
            l = line(plist1[i],add(plist1[i],plist2[i]))
            lbx = bbox(l)
            if boxoverlap2(lbx,subbox,dim3=False):
                sublist.append(l)
            else:
                glist.append(l)
            bar2.addElement(l)

        dd2 = pygletDraw()
        dd2.linecolor = [127,79,63]
        dd2.draw(glist)
        dd2.linecolor = 'white'
        dd2.draw_bbox(subbox,dim3=False)
        dd2.linecolor = 'aqua'
        dd2.draw(sublist)
        dd2.linecolor = 'red'
        for l in sublist:
            dd2.draw_bbox(l,dim3=False)
        dd2.display()

        bar2.updateTree()
        # print("bar2: ",bar2)
        # assert False

        print ("bar2 depth: ",bar2.depth)
        slist2 = bar2.getElements(subbox)
        assert len(slist2) == len(sublist)
        for l in slist2:
            assert l in sublist
 
    @pytest.mark.visual
    def test_octree(self):
        if not VISUALTEST:
            pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

        dd = pygletDraw()
        dd.linecolor = 'white'
        dd.cameradist = 35

        # make an octree
        foo =NTree()
        # add the test boxes (lines) as elements

        for l in self.glist:
            foo.addElement(l)

#        for i in range(len(rp)):
#            foo.addElement(line(rp[i],rp2[i]))
    
        foo.updateTree()

        # draw the contents of the geometry list as lines
        dd.draw(self.glist)

        # draw the contents of the geometry list as bounding boxes

        dd.linecolor = 'red'
        for b in self.glist:
            dd.draw_bbox(b,dim3=True)
        dd.display()

        bigbox = bbox(self.glist)
        # get full contents of tree
        contents = foo.getElements(bigbox)
        print("full contents: ",vstr(contents))
        assert len(contents) == len(self.glist)
        assert len(list(filter(lambda x: not x in self.glist,
                               contents))) == 0

        # retrieve the elements that overlap with self.b10

        contents = foo.getElements(self.b10)
        print("selected contents: ",vstr(contents))

        assert len(contents) == 5
        cexpect = [self.b2, self.b10, self.b20, self.b21, self.b22]
        for c in contents:
            assert c in cexpect

        # retrieve the elements that overlap with self.b1

        contents = foo.getElements(self.b1)
        print("selected contents: ",vstr(contents))

        assert len(contents) == 2
        cexpect = [self.b1, self.b20 ]
        for c in contents:
            assert c in cexpect

        # retrieve the elements that overlap with self.b4

        contents = foo.getElements(self.b4)
        print("selected contents: ",vstr(contents))

        assert len(contents) == 2
        cexpect = [self.b4, self.b21 ]
        for c in contents:
            assert c in cexpect

        # retrieve the elements that overlap with self.b7

        contents = foo.getElements(self.b7)
        print("selected contents: ",vstr(contents))

        assert len(contents) == 2
        cexpect = [self.b7, self.b22 ]
        for c in contents:
            assert c in cexpect
            
        
        foo2 = NTree(mindim=4.0)
        # make 1000 lines in space, keep track of which fall inside a
        # sub-region, then try to retrieve those with our tree,

        bigbox = [[-15,-15,-15,1],[15,15,15,1]]
        deltabox = [[-10,-10,-10,1],[10,10,10,1]]
        subbox = [[10,-10,5,1],[20,0,15,1]]
        glist=[]
        sublist=[]
        
        dim = 500
        plist1 = randomPoints(bigbox,dim)
        plist2 = randomPoints(deltabox,dim)
        
        for i in range(dim):
            l = line(plist1[i],add(plist1[i],plist2[i]))
            lbx = bbox(l)
            if boxoverlap2(lbx,subbox):
                sublist.append(l)
            else:
                glist.append(l)
            foo2.addElement(l)

        dd2 = pygletDraw()
        dd2.linecolor = [127,79,63]
        dd2.draw(glist)
        dd2.linecolor = 'white'
        dd2.draw_bbox(subbox,dim3=True)
        dd2.linecolor = 'aqua'
        dd2.draw(sublist)
        dd2.linecolor = 'red'
        for l in sublist:
            dd2.draw_bbox(l,dim3=True)
        dd2.display()

        foo2.updateTree()
            
        print ("foo2 depth: ",foo2.depth)
        slist2 = foo2.getElements(subbox)
        assert len(slist2) == len(sublist)
        for l in slist2:
            assert l in sublist
        
  



## yapCAD geometry-generating superclass
## =====================================

## Copyright (c) 2020 Richard W. DeVaul
## Copyright (c) 2020 yapCAD contributors
## All rights reserved

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""object-oriented computational geometry figure classes for **yapCAD**

===============
Overview
===============

The ``yapcad.geometry`` module provides the ``Geometry`` class. At its
most basic level, the ``Geometry`` class wrapps ``yapcad.geom``
figures and caches properties, such as the figure's bounding box,
length, center, etc., speeding certain computational geometry
operations.

Why would I use ``yapcad.geometry`` vs. ``yapcad.geom``
=======================================================

You might perfer the convenience and simplicity of the object-oriented
interface as a way to access the underlying power of the
``yapcad.geom`` and associated modules.  In addition, in many cases it
is actually more efficient to use the ``Geometry`` class wrappers for
figures rather than the ``yapcad.geom`` figures themselves because of
the memoizing and caching feautres provided.

For examle, Geometry class provides wrappers around the
``yapcad.geom3d poly2surface()`` function for the triangulation of
figures into triangle mesh surfaces, which is cached. This, in turn,
provides the foundations for extrusions and the construction of 3D
triangluated objects.

Likewise, for compound figures above a certain complexity threshold,
the Geometry class implements an internal quadtree decomposition of
the figure to speed intersection testing.  This is done in a lazy way,
which is to say that the quadtree is constructed the first time
intersection calculation is requested, and persists for as long as the
figure's geometry remains unchanged.  *NOTE: Fixeme, quadtree-based
intersection not yet implemented*

object oriented vs. functional approch
--------------------------------------

The ``yapcad.geom`` module generally takes a functional programming
approach, minimizing side effects.  This comes at the cost of some
redundant computation, as the representation of ``yapcad.geom``
figures are generally quite minimal, and properties such as centers,
bounding boxes, lengths, **etc.**, must be recomputed each time they
are requested, unless the user has made their own explicit cache. 

By contrast, the ``Geometry`` class caches these properties, and can
provide higher efficiency for certain types of complex operations.  In
addition, derived classes provide representations and functionality
absent from the underlying functional representations, such as
"growable" polygons, boolean operations, *etc.*.

yapcad.geometry conveneince functions
=====================================

Convenience functions, such as ``Line()``, ``Arc()``, *etc.*, operate
analogously to their uncapitalized counterparts in ``yapcad.geom``,
creating ``Geometry`` class instances wrapping the corresponding
figure.

"""

from copy import deepcopy
from yapcad.geom import *
from yapcad.geom_util import *
from yapcad.geom3d import *

class Geometry:
    """generalized computational geometry base class, also acts as a
    wrapper around yapcad.geom elements.  

    Using Geometry subclass qinstances can be more effiient than
    workng with the corresponding yapcad.geom elements, as the
    Geometry instance uses lazy evaluation and caching to speed
    repeated evaluations of various properties.

    """

    def __repr__(self):
        return f"Geometry({self.__elem})"

    def __init__(self,a=False):
        self.__update=True # do we need to update geom?
        self.__elem=[] # internal geometry
        self.__length=0.0
        self.__sampleable = False
        self.__intersectable = False
        self.__continuous = False
        self.__closed = False
        self.__center = None
        self.__bbox = None
        self.__surface = None # surface representation
        self.__surface_ang = -1 # surface parameter
        self.__surface_len = -1 # surface parameter
        self.__derived = False  # set to true for derived geometry subclasses
        if a != False:
            if ispoint(a):
                self.__elem=  deepcopy(a) 
                self.__sampleable=True
            elif isline(a) or isarc(a):
                self.__elem= deepcopy(a)
                self.__sampleable=True
                self.__intersectable=True
                self.__continuous=True
                if iscircle(a):
                    self.__closed = True
            elif ispoly(a):
                self.__elem=deepcopy(a)
                self.__sampleable=True
                self.__intersectable=True
                self.__continuous=True
                self.__closed=ispolygon(a)
            elif isgeomlist(a):
                self.__elem=deepcopy(a)
                self.__sampleable=True
                self.__intersectable=True
                self.__continuous=iscontinuousgeomlist(a)
                self.__closed=isclosedgeomlist(a)

            elif isinstance(a,Geometry):
                self.__elem = deepcopy(a.elem)
                self.__sampleable = a.issampleable()
                self.__intersectable= a.isintersectable()
                self.__continuous= a.iscontinuous()
                self.__closed = a.isclosed()
            else:
                raise ValueError(f'bad argument to Geometry class constructor: {a}')


    @property
    def sampleable(self):
        return self.__sampleable

    def _setSampleable(self,bln):
        self.__sampleable=bln

    def issampleable(self):
        """is the figure sampleable?"""
        return self.__sampleable

    @property
    def intersectable(self):
        return self.__intersectable

    def isintersectable(self):
        """is the figure intersectable?"""
        return self.__intersectable

    @property
    def continuous(self):
        return self.__continuous

    def iscontinuous(self):
        """is the figure C0 continuous over the interval [0,1]"""
        return self.__continuous

    @property
    def closed(self):
        return self.__closed

    def _setClosed(self,bln):
        self.__closed=bln

    def isclosed(self):
        """is the figure C0 continuous over the [0,1] interval and
        is ``self.sample(0.0)`` within epsilon of ``self.sample(1.0)``
        """
        return self.__closed

    @property
    def derived(self):
        return self.__derived

    def _setDerived(self,bln):
        self.__derived = bln

    def isderived(self):
        """is this an instance of a derived geometry subclass that
        computes self.geom from self.elem?  Set only in constructors
        for derived geometry subclasses
        """
        return self.__derived

    @property
    def length(self):
        """return length of figure"""
        if self.update:
            self._updateInternals()
        return self.__length

    def _setLength(self,l):
        self.__length=l

    @property
    def center(self):
        """return center of figure"""
        if self.update:
            self._updateInternals()
        return self.__center

    def _setCenter(self,c):
        self.__center=c

    @property
    def bbox(self):
        """return 3D bounding box of figure"""
        if self.update:
            self._updateInternals()
        return self.__bbox

    def _setBbox(self,bbx):
        self.__bbox=bbx

    def isinsideXY(self,p):
        """determine if a point is inside a figure.  In the case of non-closed
        figures, such as lines, determine if the point lies within
        epsilon of one of the lines of the figure.
        """
        if self.update:
            self._updateInternals()
        return isinsideXY(self.geom,p)


    def translate(self,delta):
        """apply a translation to figure"""
        self._setElem(translate(self.__elem,delta))
        self._setUpdate(True)

    def scale(self,sx=1.0,sy=False,sz=False,cent=point(0,0,0)):
        """apply a scaling to a figure"""
        self._setElem(scale(self.__elem,sx,sy,sz,cent))
        self._setUpdate(True)

    def rotate(self,ang,cent=point(0,0,0),axis=point(0,0,1.0)):
        """apply a rotation to a figure"""
        self._setElem(rotate(self.__elem,ang,cent,axis))
        self._setUpdate(True)

    def mirror(self,plane,keepSign=True):
        """apply a mirror operation to a figure.  Currently, the following
        values of "plane" are allowed: 'xz', 'yz', xy'.  Generalized
        arbitrary reflection plane specification will be added in the
        future.

        If ``keepSign == True`` (default) the sign of the area will be
        maintained, meaning that if ``mirror`` is applied to a
        right-handed closed figure (a figure with positive area) the
        resulting mirrored figure will also have positive area.  This
        is probably what you want, unless you are specifically turning
        a face into a hole, or vice-versa.

        """
        nelm = mirror(self.__elem,plane)
        if keepSign:
            nelm = reverseGeomList(nelm)
        self._setElem(nelm)
        self._setUpdate(True)

    def transform(self,m):
        """apply an arbitrary transformation to a figure, as specified by a
        transformation matrix.
        """
        self._setElem(transform(self.elem,m))
        self._setUpdate(True)
        

    # one underscore to make this easily overridable in subclasses
    def _updateInternals(self):
        """update internals: set basic attributes based on geom"""
        if self.update:
            self._setUpdate(False)
            if self.__elem == []:
                self.__length = 0.0
                self.__center = None
                self.__bbox = None
            else:
                self.__length = length(self.__elem)
                self.__center = center(self.__elem)
                self.__bbox = bbox(self.__elem)

        return

    ## more properties
    
    @property
    def update(self):
        return self.__update

    def _setUpdate(self,bln):
        """method to set status of __update, accessible from derived classes"""
        self.__update = bln

    @update.setter
    def update(self,bln):
        """property to flag update.  Ignores right-hand side of expression, always sets update flag to True"""
        self._setUpdate(True)

    @property 
    def elem(self):
        return self.__elem

    def _setElem(self,e):
        self.__elem = e

    @elem.setter
    def elem(self,e):
        self._setElem(e)

    @property
    def geom(self):
        """return yapcad.geom representation of figure"""
        if self.update:
            self._updateInternals()
        return deepcopy(self.__elem)

    def sample(self,u):
        """If the figure is sampleable, given a parameter u, return the point
        on the figure corresponding to the specified sampling
        parameter.

        """
        if not self.__sampleable:
            raise ValueError('figure is not sampleable')
        if self.update:
            self._updateInternals()
            
        gl = self.geom

        if len(gl) == 1:
            return sample(gl[0],u)
        else:
            return sample(gl,u)

    def unsample(self,p):
        """
        Invert the sampling operation: return the parameter corresponding
        to the closest point on the figure to p as long as the distance is
        less than epsilon.
        
        """
        if not self.__sampleable:
             raise ValueError('figure is not sampleable, or unsampleable for that matter')
        if self.update:
            self._updateInternals()

        gl = self.geom

        if len(gl) == 1:
            return unsample(gl[0],p)
        else:
            return unsample(gl,p)

    def segment(self,u1,u2,reverse=False):
        gl = self.geom
        if gl == []:
            raise ValueError('empty Boolean, segment not defined')
        return segmentgeomlist(gl,u1,u2,closed=self.closed,reverse=reverse)

    def intersectXY(self,g,inside=True,params=False):
        """given two XY-coplanar figures, this figure and ``g``,
        calculate the intersection of these two figures, and return a
        list of intersection points, or False if none.  

        ``g`` must be an instance of ``IntersectGeometry``, or a
        ``yapcad.geom`` figure.

        If ``inside == True``, only return intersections that are
        within the ``0 <= u <= 1.0`` interval for both figures.  If
        ``params == True``, instead of returning a list of points,
        return two lists corresponding to the sampling parameter value
        of the intersections corresponding to each figure.

        """
        if not self.isintersectable():
            raise ValueError(f'this instance {self} not intersectable')
        if self.update:
            self._updateInternals()
        if isinstance(g,Geometry):
            if g.isintersectable():
                g = g.geom
            else:
                raise ValueError(f'Geometry object {g} not intersectable')
        elif isgeomlist([g]):
            pass
        else:
            raise ValueError(f'bad thing passed to intersectXY: {g}')
        
        return intersectXY(self.geom,g,inside,params)

    
    def surface(self,minang = 5.0, minlen = 0.5):
        """
        triangulate a closed polyline or geometry list, return a surface.
        the ``minang`` parameter specifies a minimum angular resolution
        for sampling arcs, and ``minleng`` specifies a minimum distance
        between sampled points. 
        ``surface = ['surface',vertices,normals,faces]``, where:
            ``vertices`` is a list of ``yapcad.geom`` points,
            ``normals`` is a list of ``yapcad.geom`` points of the same length as ``vertices``,
            and ``faces`` is the list of faces, which is to say lists of three indices that 
            refer to the vertices of the triangle that represents each face.
        """
        if not self.isclosed():
            raise ValueError("non-closed figure has no surface representation")
        if self.update:
            self._updateInternals()
        if (self.__surface and
            close(self.__surface_ang,minang) and
            close(self.__surface_len,minlen)):
            return self.__surface
        self.__surface_ang = minang
        self.__surface_len = minlen
        geo = self.geom
        if len(geo) == 0:
            return []
        if ispolygon(geo):
            ply = geo
            holes = []
        else:
            ply, holes = geomlist2poly_with_holes(geo, minang, minlen)

        self.__surface,bnd = poly2surface(ply, holepolys=holes, checkclosed=False)

        return self.__surface

## Utility functions

def Point(x=False,y=False,z=False,w=False):
    return Geometry(point(x,y,z,w))

def Line(p1,p2=False):
    return Geometry(line(p1,p2))

def Arc(c,rp=False,sn=False,e=False,n=False,samplereverse=False):
    return Geometry(arc(c,rp,sn,e,n,samplereverse))

# def Poly(*args):
#     ply = poly(*args)
#     # print(f'ply: {ply}')
#     return Geometry(ply)

def Figure(*args):
    return Geometry(list(*args))


                    



import uuid

from yapcad.geom import point
from yapcad.geom3d import poly2surfaceXY, solid, surface
from yapcad.metadata import (
    ensure_solid_id,
    ensure_surface_id,
    get_solid_metadata,
    get_surface_metadata,
    set_solid_context,
    set_solid_metadata,
    set_surface_metadata,
    set_surface_origin,
    set_surface_units,
)


def test_surface_metadata_roundtrip():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]
    surf, _ = poly2surfaceXY(poly)

    assert get_surface_metadata(surf) == {}

    ensure_surface_id(surf)
    meta = get_surface_metadata(surf)
    assert 'id' in meta

    set_surface_units(surf, 'mm')
    set_surface_origin(surf, 'parametric:test')
    meta = get_surface_metadata(surf)
    assert meta['units'] == 'mm'
    assert meta['origin'] == 'parametric:test'

    custom = {'foo': 'bar'}
    set_surface_metadata(surf, custom)
    assert get_surface_metadata(surf) is custom


def test_surface_constructor_with_metadata():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]
    surf, _ = poly2surfaceXY(poly, holepolys=[], minlen=0.5)
    metadata = {'id': '1234'}
    surf_with_meta = surface(surf[1], surf[2], surf[3], surf[4], surf[5], metadata)
    assert get_surface_metadata(surf_with_meta) == metadata


def test_solid_metadata_helpers():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]
    surf, _ = poly2surfaceXY(poly)
    sld = solid([surf])

    assert get_solid_metadata(sld) == {}

    ensure_solid_id(sld)
    meta = get_solid_metadata(sld)
    assert 'id' in meta

    set_solid_context(sld, {'solver': 'dummy'})
    assert get_solid_metadata(sld)['context'] == {'solver': 'dummy'}

    custom = {'id': str(uuid.uuid4())}
    set_solid_metadata(sld, custom)
    assert get_solid_metadata(sld) is custom



#!/usr/bin/env python3
"""Run CAM-oriented linting on triangular meshes.

This script chains together three pragmatic checks that mirror the
downstream tools typically used in fabrication workflows:

1. ``admesh`` for statistics about the STL (holes, connected components,
   facet orientation, volume, bounding box, etc.).
2. ``meshfix`` for geometry healing (fills small holes, removes dangling
   triangles, produces a cleaned STL for further work).
3. A slicer front-end such as ``prusa-slicer``/``prusa-slicer-console``/
   ``slic3r``/``curaengine`` to prove the model slices successfully.

Each step is optional – if a tool is not installed the script will emit a
warning and carry on.  The goal is to provide a quick report that correlates
better with real-world CAM robustness than the strict, topological
``issolidclosed`` test.

Example
=======

.. code-block:: bash

    PYTHONPATH=src python examples/solid_boolean_demo.py \
        --mode stl --operation difference --shapes box_hole \
        --output box_hole_difference

    python tools/validate_mesh.py box_hole_difference.stl \
        --workdir build/mesh_checks

The command above keeps all intermediate files (repaired STL, generated
G-code, log snippets) under ``build/mesh_checks`` so they can be inspected
or archived alongside the STEP export.
"""

from __future__ import annotations

import argparse
import json
import os
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Iterable, Optional


def _run_pymeshfix(stl: Path, output: Path) -> Optional[dict[str, object]]:
    """Attempt to repair ``stl`` using the PyMeshFix Python bindings.

    Returns a summary dictionary if successful, otherwise ``None`` when the
    dependency is missing or the mesh could not be loaded.
    """

    try:
        import pymeshfix  # type: ignore[import-untyped]
        import trimesh  # type: ignore[import-untyped]
    except ImportError:
        return None

    try:
        mesh = trimesh.load(str(stl), force='mesh')
    except Exception as exc:  # pragma: no cover - depends on trimesh internals
        return {
            "returncode": 1,
            "stdout": "",
            "stderr": f"Failed to load STL with trimesh: {exc}",
        }

    if not isinstance(mesh, trimesh.Trimesh) or mesh.vertices.size == 0:
        return {
            "returncode": 1,
            "stdout": "",
            "stderr": "Input did not contain a valid triangular mesh",
        }

    fixer = pymeshfix.MeshFix(mesh.vertices, mesh.faces)
    fixer.repair(verbose=False)
    try:
        fixer.write_stl(str(output))
    except AttributeError:
        # Older PyMeshFix releases expose the repaired data via ``v``/``f``.
        repaired = trimesh.Trimesh(fixer.v, fixer.f, process=False)
        repaired.export(str(output))
    return {
        "returncode": 0,
        "stdout": "PyMeshFix repair completed",
        "stderr": "",
        "output": str(output),
    }


ToolName = str


def which(names: Iterable[ToolName]) -> Optional[Path]:
    """Return the first executable in *names* that exists on PATH."""

    for name in names:
        path = shutil.which(name)
        if path:
            return Path(path)
    return None


def run(cmd: Iterable[str], *, cwd: Optional[Path] = None) -> subprocess.CompletedProcess:
    """Execute *cmd*, capturing stdout/stderr for later inspection."""

    result = subprocess.run(
        list(cmd),
        cwd=str(cwd) if cwd else None,
        check=False,
        text=True,
        encoding="utf-8",
        errors="replace",
        capture_output=True,
    )
    return result


def parse_admesh_report(output: str) -> dict[str, str]:
    """Extract a few useful numbers from ``admesh`` output."""

    summary: dict[str, str] = {}
    for line in output.splitlines():
        line = line.strip()
        if not line:
            continue
        for prefix in (
            "Number of facets:",
            "Volume:",
            "Number of shells:",
            "Number of pieces of geometry:",
            "Number of edges (with 2 facets):",
            "Number of edges (with 1 facet):",
        ):
            if line.startswith(prefix):
                summary[prefix.rstrip(':')] = line.split(':', 1)[1].strip()
    return summary


def slicer_command(output_dir: Path, stl: Path) -> Optional[list[str]]:
    """Return a slicer command if a supported engine is installed."""

    slicers: list[tuple[list[str], list[str]]] = [
        (["prusa-slicer-console", "prusa-slicer", "PrusaSlicer"], ["--help"]),
        (["slic3r"], ["--help"]),
        (["curaengine"], ["--help"]),
    ]

    for names, probe_args in slicers:
        binary = which(names)
        if not binary:
            continue

        # Make sure the executable is functional by running the probe command.
        probe = run([str(binary), *probe_args])
        if probe.returncode != 0:
            continue

        gcode = output_dir / (stl.stem + ".gcode")

        if binary.name.startswith("curaengine"):
            profile = os.environ.get("CURA_PROFILE")
            if not profile:
                continue
            return [str(binary), "slice", "-v", "-o", str(gcode), "-j", profile, "-l", str(stl)]

        if binary.name.startswith("slic3r"):
            return [
                str(binary),
                str(stl),
                "--output",
                str(gcode),
                "--loglevel",
                "0",
            ]

        # Default for PrusaSlicer variants.
        return [
            str(binary),
            "--export-gcode",
            str(stl),
            "--output",
            str(gcode),
        ]

    return None


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(
        description="Run mesh linting/healing/slicing checks in one pass.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("stl", type=Path, help="Path to the STL file to validate")
    parser.add_argument(
        "--workdir",
        type=Path,
        help="Directory to store intermediate files (default: temporary directory).",
    )
    parser.add_argument(
        "--json",
        action="store_true",
        help="Emit a machine-readable JSON summary instead of human text.",
    )

    args = parser.parse_args(argv)
    stl_path = args.stl

    if not stl_path.exists():
        parser.error(f"STL file not found: {stl_path}")

    if not stl_path.suffix.lower() == ".stl":
        parser.error("The input file must be an STL")

    if args.workdir:
        workdir = args.workdir
        workdir.mkdir(parents=True, exist_ok=True)
        cleanup = False
    else:
        workdir = Path(tempfile.mkdtemp(prefix="yapcad-meshcheck-"))
        cleanup = True

    summary: dict[str, object] = {
        "input": str(stl_path.resolve()),
        "workdir": str(workdir.resolve()),
        "admesh": None,
        "meshfix": None,
        "slicer": None,
        "notes": [],
    }

    # 1. admesh ------------------------------------------
    admesh_bin = which(["admesh"])
    if admesh_bin:
        repaired_stl = workdir / f"{stl_path.stem}-admesh.stl"
        cmd = [str(admesh_bin), str(stl_path)]
        result = run(cmd)
        admesh_info = {
            "returncode": result.returncode,
            "stdout": result.stdout,
            "stderr": result.stderr,
        }
        admesh_info.update(parse_admesh_report(result.stdout))
        summary["admesh"] = admesh_info
    else:
        summary["notes"].append("admesh not found on PATH; skipping mesh statistics")

    # 2. meshfix -----------------------------------------
    meshfix_bin = which(["meshfix"])
    if meshfix_bin:
        meshfix_out = workdir / f"{stl_path.stem}-meshfix.stl"
        cmd = [str(meshfix_bin), str(stl_path), "-o", str(meshfix_out)]
        result = run(cmd)
        summary["meshfix"] = {
            "returncode": result.returncode,
            "stdout": result.stdout,
            "stderr": result.stderr,
            "output": str(meshfix_out),
        }
        if result.returncode != 0:
            summary["notes"].append("meshfix reported an error; see stderr for details")
    else:
        meshfix_out = workdir / f"{stl_path.stem}-pymeshfix.stl"
        pymeshfix_info = _run_pymeshfix(stl_path, meshfix_out)
        if pymeshfix_info is not None:
            summary["meshfix"] = pymeshfix_info
            if pymeshfix_info.get("returncode") != 0:
                summary["notes"].append("PyMeshFix reported an issue; see stderr for details")
        else:
            summary["notes"].append("meshfix not found and PyMeshFix unavailable; skipping repair step")

    # 3. slicer ------------------------------------------
    slicer_cmd = slicer_command(workdir, stl_path)
    if slicer_cmd:
        result = run(slicer_cmd)
        summary["slicer"] = {
            "command": slicer_cmd,
            "returncode": result.returncode,
            "stdout": result.stdout,
            "stderr": result.stderr,
        }
        if result.returncode != 0:
            summary["notes"].append("Slicer failed – inspect stdout/stderr for diagnostics")
    else:
        summary["notes"].append(
            "No supported slicer (prusa-slicer, slic3r, curaengine) found; skipping G-code export"
        )

    if args.json:
        print(json.dumps(summary, indent=2))
    else:
        lines = [f"Mesh validation report for {stl_path}"]
        lines.append("Output directory: " + str(workdir))
        lines.append("")

        if summary["admesh"]:
            lines.append("admesh summary:")
            admesh_info = summary["admesh"]  # type: ignore[assignment]
            for key in (
                "Number of facets",
                "Number of facets connected to:",
                "Number of edges (with 1 facet)",
                "Number of edges (with 2 facets)",
                "Number of shells",
                "Volume",
            ):
                value = admesh_info.get(key)
                if value is not None:
                    lines.append(f"  {key}: {value}")
            lines.append("")
        else:
            lines.append("admesh: skipped")
            lines.append("")

        if summary["meshfix"]:
            meshfix_info = summary["meshfix"]  # type: ignore[assignment]
            lines.append("meshfix: return code %s" % meshfix_info["returncode"])
            lines.append("  output: %s" % meshfix_info["output"])
            lines.append("")
        else:
            lines.append("meshfix: skipped")
            lines.append("")

        if summary["slicer"]:
            slicer_info = summary["slicer"]  # type: ignore[assignment]
            cmd_str = " ".join(str(x) for x in slicer_info.get("command", []))
            lines.append("slicer command: " + cmd_str)
            lines.append("  return code: %s" % slicer_info["returncode"])
            lines.append("")
        else:
            lines.append("slicer: skipped")
            lines.append("")

        if summary["notes"]:
            lines.append("Notes:")
            for note in summary["notes"]:
                lines.append("  - " + note)
        else:
            lines.append("No warnings.")

        print("\n".join(lines))

    if cleanup:
        shutil.rmtree(workdir, ignore_errors=True)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""Diagnose individual yapCAD primitives using trimesh helpers."""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from typing import Any, Dict

from mesh_diagnostics import diagnose_mesh, export_boundary

from yapcad.geom import point
from yapcad.geom3d_util import prism, sphere, conic, tube
from yapcad.boolean.trimesh_engine import _solid_to_mesh


def _build_primitive(name: str, params: Dict[str, Any]):
    if name == 'sphere':
        diameter = float(params.get('diameter', 2.0))
        depth = int(params.get('depth', 2))
        return sphere(diameter, depth=depth)
    if name == 'prism':
        length = float(params.get('length', 2.0))
        width = float(params.get('width', 2.0))
        height = float(params.get('height', 2.0))
        return prism(length, width, height)
    if name == 'conic':
        baser = float(params.get('base_radius', 1.0))
        topr = float(params.get('top_radius', 0.5))
        height = float(params.get('height', 2.0))
        angr = float(params.get('angular_resolution', 10.0))
        return conic(baser, topr, height, center=params.get('center', point(0, 0, 0)), angr=angr)
    if name == 'tube':
        outer_diameter = float(params.get('outer_diameter', 3.0))
        wall_thickness = float(params.get('wall_thickness', 0.5))
        length = float(params.get('length', 4.0))
        include_caps = bool(params.get('include_caps', True))
        base_point = params.get('base_point')
        if base_point is not None:
            base_point = point(*base_point)
        return tube(outer_diameter=outer_diameter, wall_thickness=wall_thickness,
                    length=length, base_point=base_point, include_caps=include_caps)
    raise ValueError(f'unsupported primitive {name!r}')


def main():
    parser = argparse.ArgumentParser(description='Diagnose individual yapCAD primitives.')
    parser.add_argument('--primitive', choices=['sphere', 'prism', 'conic', 'tube'], required=True)
    parser.add_argument('--params', help='JSON object with primitive parameters (e.g. {"diameter": 3.0})')
    parser.add_argument('--boundary-output', type=Path, help='optional path to export boundary edges (PLY)')
    parser.add_argument('--json', type=Path, help='write diagnostics JSON to path')
    parser.add_argument('--export-stl', type=Path, help='optional path to export the mesh as STL')
    args = parser.parse_args()

    params: Dict[str, Any] = {}
    if args.params:
        try:
            params = json.loads(args.params)
        except json.JSONDecodeError as exc:
            raise SystemExit(f'Failed to parse --params JSON: {exc}') from exc

    solid = _build_primitive(args.primitive, params)
    mesh = _solid_to_mesh(solid)
    diagnostics = diagnose_mesh(mesh)

    if args.boundary_output:
        export_boundary(mesh, args.boundary_output)

    if args.export_stl:
        args.export_stl.parent.mkdir(parents=True, exist_ok=True)
        mesh.export(str(args.export_stl))

    print(json.dumps(diagnostics.to_dict(), indent=2))

    if args.json:
        args.json.parent.mkdir(parents=True, exist_ok=True)
        args.json.write_text(json.dumps(diagnostics.to_dict(), indent=2))


if __name__ == '__main__':
    main()



import io
import struct

from yapcad.geom import point
from yapcad.geom3d import poly2surfaceXY
from yapcad.io.stl import write_stl


def _make_surface():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]
    surf, _ = poly2surfaceXY(poly)
    return surf


def test_write_stl_binary(tmp_path):
    surf = _make_surface()
    path = tmp_path / 'tri.stl'
    write_stl(surf, path, binary=True, name='test')

    data = path.read_bytes()
    assert len(data) == 80 + 4 + 50  # header + count + one triangle
    assert data[0:4] == b'test'
    count = struct.unpack('<I', data[80:84])[0]
    assert count == 1


def test_write_stl_ascii():
    surf = _make_surface()
    buf = io.StringIO()
    write_stl(surf, buf, binary=False, name='ascii_test')

    text = buf.getvalue()
    assert 'solid ascii_test' in text
    assert 'facet normal' in text
    assert 'vertex' in text
    assert text.strip().endswith('endsolid ascii_test')



## drawing and computational geometry example for yapCAD
print("example3.py -- yapCAD computational geometry and DXF drawing example")
print('''
Create three circles, then compute intersections between circles and
draw tangent lines between pairs of circles.''')

from yapcad.ezdxf_drawable import *
from yapcad.geom import *

#set up DXF rendering
d=ezdxfDraw()

filename="example3-out"
print("\nOutput file name is {}.dxf".format(filename))
d.filename = filename

d.layer = 'DOCUMENTATION' # select the DOCUMENTATION layer
d.draw_text("yapCAD", point(-9,7),\
            attr={'style': 'OpenSans-Bold',
                  'height': 1.5})

d.draw_text("example3.py",
            point(-9,5.0))
d.draw_text("circles, tangents, and intersections",
            point(-9,3.5))
d.layer = False # select the default layer

a = point(5,5)
b = point(-5,5)
c = point(0,0)

circ1 = arc(a,5)
circ2 = arc(b,7.5)
circ3 = arc(c,1)

tl1 = circleCircleTangentsXY(circ1,circ2)
tl2 = circleCircleTangentsXY(circ3,circ1)

i1 = intersectXY(circ1,circ2,False)
i2 = intersectXY(circ2,circ3,False)

points = i1 + i2

d.draw(circ1)
d.draw(circ2)
d.draw(circ3)

d.draw(tl1[0])
d.draw(tl1[1])

d.draw(tl2[0])
d.draw(tl2[1])

## configure point drawing, draw on DOCUMENTATION layer
d.pointstyle='o'
d.polystyle='points'
d.layer = 'DOCUMENTATION'
d.draw(points)

d.display()



import os
import pytest

from yapcad.combine import Boolean
from yapcad.geom import arc, line, point, translate
from yapcad.geom3d import poly2surfaceXY, solid, translatesurface, surf2lines
from yapcad.geom3d_util import conic, sphere
from yapcad.geom_util import geomlist2poly, geomlist2poly_components
from yapcad.mesh import mesh_view
from yapcad.poly import Polygon, Rect

VISUALTEST = os.environ.get('VISUALTEST', 'false').lower() in ('true', '1', 'yes')
if VISUALTEST:
    from yapcad.pyglet_drawable import pygletDraw

def _strip_w(pt):
    return tuple(round(coord, 6) for coord in pt[:3])


def test_mesh_view_surface_basic():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]

    surf, _ = poly2surfaceXY(poly)

    tris = list(mesh_view(surf))
    assert len(tris) == 1

    normal, v0, v1, v2 = tris[0]
    assert normal == (0.0, 0.0, 1.0)
    assert _strip_w(surf[1][0]) in (v0, v1, v2)


def test_mesh_view_reorients_against_stored_normal():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]

    surf, _ = poly2surfaceXY(poly)
    surf[2] = [[0, 0, -1, 0] for _ in surf[2]]

    tri = next(mesh_view(surf))
    normal = tri[0]
    assert normal == (0.0, 0.0, 1.0)


def test_mesh_view_on_solid():
    poly = [
        point(0, 0),
        point(1, 0),
        point(0, 1),
        point(0, 0),
    ]

    surf, _ = poly2surfaceXY(poly)
    sld = solid([surf], [], [])

    tris = list(mesh_view(sld))
    assert len(tris) == 1


@pytest.mark.visual
def test_mesh_view_visual_normals():
    if not VISUALTEST:
        pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

    outer = [
        point(-1, -1),
        point(1, -1),
        point(1, 1),
        point(-1, 1),
        point(-1, -1),
    ]

    hole = [
        point(-0.4, -0.4),
        point(0.4, -0.4),
        point(0.4, 0.4),
        point(-0.4, 0.4),
        point(-0.4, -0.4),
    ]

    surf, _ = poly2surfaceXY(outer, holepolys=[hole])
    sld = solid([surf], [], [])

    dd = pygletDraw()

    def draw_normals(obj, color):
        dd.linecolor = color
        for normal, v0, v1, v2 in mesh_view(obj):
            cx = (v0[0] + v1[0] + v2[0]) / 3.0
            cy = (v0[1] + v1[1] + v2[1]) / 3.0
            cz = (v0[2] + v1[2] + v2[2]) / 3.0
            start = point(cx, cy, cz + 0.01)
            end = point(
                cx + normal[0] * 0.5,
                cy + normal[1] * 0.5,
                cz + normal[2] * 0.5 + 0.01,
            )
            dd.draw(line(start, end))

    dd.linecolor = 'silver'
    dd.polystyle = 'both'
    dd.draw_surface(surf)
    draw_normals(sld, 'yellow')

    shp = sphere(1.5, center=point(3, 0, 0))
    dd.linecolor = 'white'
    dd.draw(shp)
    draw_normals(shp, 'aqua')

    cone = conic(1.0, 0.5, 2.5, center=point(-3, 0, 0))
    dd.linecolor = 'green'
    dd.draw(cone)
    draw_normals(cone, 'blue')

    dd.display()


@pytest.mark.visual
def test_mesh_view_visual_multi_component_tessellation():
    if not VISUALTEST:
        pytest.skip("Visual tests disabled (set VISUALTEST=true to enable)")

    dd = pygletDraw()
    dd.polystyle = 'both'

    multi_outer = [
        point(-4, -2),
        point(0, -2),
        point(2, 0),
        point(0, 2),
        point(-4, 2),
        point(-4, -2),
    ]

    hole_one = [
        point(-3.0, -1.0),
        point(-2.0, -1.0),
        point(-2.0, 0.0),
        point(-3.0, 0.0),
        point(-3.0, -1.0),
    ]

    hole_two = [
        point(-1.5, 0.5),
        point(-0.5, 0.5),
        point(-0.5, 1.5),
        point(-1.5, 1.5),
        point(-1.5, 0.5),
    ]

    multi_surface, _ = poly2surfaceXY(multi_outer, holepolys=[hole_one, hole_two])
    dd.linecolor = 'orange'
    dd.draw_surface(multi_surface)

    outer_a = Rect(4, 4)
    outer_b = Rect(2, 2, center=point(5.5, 0))
    union_shape = Boolean('union', [outer_a, outer_b])
    components = geomlist2poly_components(union_shape.geom)

    palette = ['cyan', 'magenta', 'yellow', 'lime']

    for idx, (outer_poly, holes) in enumerate(components):
        surf, _ = poly2surfaceXY(outer_poly, holepolys=holes)
        shifted = translatesurface(surf, point(6, idx * 3.5, 0))
        dd.linecolor = palette[idx % len(palette)]
        dd.draw_surface(shifted)

    def _circle_polygon(cx, cy, radius, segments=48):
        arc_geom = [arc(point(cx, cy), radius)]
        sampled = geomlist2poly(arc_geom, minang=360.0 / segments)
        return Polygon(sampled)

    base_square = Rect(6, 6)
    circle_a = _circle_polygon(-1.6, -0.8, 1.2)
    circle_b = _circle_polygon(1.6, 0.8, 1.0)
    carved_first = Boolean('difference', [base_square, circle_a])
    carved = Boolean('difference', [carved_first, circle_b])
    carved_components = geomlist2poly_components(carved.geom)

    if carved_components:
        carved_outer, carved_holes = carved_components[0]
        carved_surface, _ = poly2surfaceXY(carved_outer, holepolys=carved_holes)
        carved_surface = translatesurface(carved_surface, point(-6, -4, 0))
        dd.linecolor = 'lightsteelblue'
        dd.draw_surface(carved_surface)

        edge_lines = surf2lines(carved_surface)
        edge_lines = translate(edge_lines, point(0, 0, 0.15))
        dd.linecolor = 'yellow'
        dd.draw(edge_lines)

    dd.display()
